from .base_summary import Summary
from .classification_summary import ClassificationSummary
from .regression_summary import RegressionSummary
from .reconstruction_summary import ReconstructionSummary
from .multi_model_summary import MultiModelSummary


import torch as t
import torch.nn as nn


class Reshape(nn.Module):

    def __init__(self, *args):
        super().__init__()
        self.shape = args

    def forward(self, x):
        # keep batch size
        return x.view(x.shape[0], *self.shape)


class KerasLikeLSTM(nn.Module):

    def __init__(self, input_shape, output, return_sequence=False, sequential=False):
        super().__init__()
        self.features = input_shape[1] if isinstance(input_shape, tuple) else 1
        self.timesteps = input_shape[0] if isinstance(input_shape, tuple) else input_shape
        self.output = output
        self.lstm = nn.LSTM(self.features, self.output, batch_first=True)
        self.return_sequence = return_sequence
        self.sequential = sequential
        self.hidden = None
        self.cell = None

    def forward(self, x):
        if not self.sequential or self.hidden is None:
            self.hidden = t.zeros(1, x.size(0), self.output).to(next(self.parameters()).device)

        self.cell = t.zeros(1, x.size(0), self.output).to(next(self.parameters()).device)
        out, self.hidden = self.lstm(x, (self.hidden, self.cell))

        if self.return_sequence:
            return out
        else:
            # only return last dimension
            return out[:, -1]


class ResidualLayer(nn.Module):

    def __init__(self, input_size, output_size, activation):
        super().__init__()
        self.l1 = nn.Linear(input_size, output_size)
        self.l2 = nn.Linear(output_size, output_size)
        self.activation = activation

    def forward(self, x):
        x1 = self.activation(self.l1(x))
        x2 = self.activation(self.l2(x1))
        return x1 + x2



from unittest import TestCase

import torch as t

from pandas_ml_utils.pytorch.layers import KerasLikeLSTM


class TestPytorchLayer(TestCase):

    def test_lstm_layer(self):
        data = t.rand((1, 10, 2))
        lstm = KerasLikeLSTM((10, 2), 2, return_sequence=False)
        lstm2 = KerasLikeLSTM((10, 2), 2, return_sequence=True)

        self.assertEqual((1, 2), lstm(data).shape)
        self.assertEqual((1, 10, 2), lstm2(data).shape)




from typing import Tuple, List, Callable

import pandas as pd
import numpy as np
import logging

from pandas_ml_common import get_pandas_object, Typing
from pandas_ml_common.decorator import MultiFrameDecorator
from pandas_ml_common.utils import intersection_of_index, loc_if_not_none
from pandas_ml_common.utils.callable_utils import call_if_not_none

_log = logging.getLogger(__name__)


def extract_feature_labels_weights(
        df: Typing.PatchedDataFrame,
        features_and_labels,
        **kwargs) -> Tuple[Tuple[pd.DataFrame, int], pd.DataFrame, pd.DataFrame, pd.Series, pd.Series]:
    _, features, targets = extract_features(df, features_and_labels, **kwargs)
    labels = get_pandas_object(df, features_and_labels.labels, **kwargs).dropna()
    sample_weights = call_if_not_none(get_pandas_object(df, features_and_labels.sample_weights, **kwargs), 'dropna')
    gross_loss = call_if_not_none(get_pandas_object(df, features_and_labels.gross_loss, **kwargs), 'dropna')

    if features_and_labels.label_type is not None:
        labels = labels.astype(features_and_labels.label_type)

    # do some sanity check for any non numeric values in any of the data frames
    for frame in [features, labels, targets, sample_weights, gross_loss]:
        if frame is not None:
            max = frame._.max()

            if np.isscalar(max) and np.isinf(max):
                _log.warning("features containing infinit number\n", frame[frame.apply(lambda r: np.isinf(r.values).any(), axis=1)])
                frame.replace([np.inf, -np.inf], np.nan, inplace=True)
                frame.dropna(inplace=True)

    # now get the common index and return the filtered data frames
    common_index = intersection_of_index(features, labels, targets, sample_weights, gross_loss)

    return (
        (
            tuple([f.loc[common_index] for f in features]) if isinstance(features, tuple) else features.loc[common_index],
            len(df) - len(features) + 1
        ),
        labels.loc[common_index],
        loc_if_not_none(targets, common_index),
        loc_if_not_none(sample_weights, common_index),
        loc_if_not_none(gross_loss, common_index)
    )


def extract_features(df: pd.DataFrame, features_and_labels, **kwargs) -> Tuple[List, pd.DataFrame, pd.DataFrame]:
    if isinstance(features_and_labels.features, tuple):
        # allow multiple feature sets i.e. for multi input layered networks
        features = MultiFrameDecorator([get_pandas_object(df, f, **kwargs).dropna() for f in features_and_labels.features], True)
    else:
        features = get_pandas_object(df, features_and_labels.features, **kwargs).dropna()

    targets = call_if_not_none(get_pandas_object(df, features_and_labels.targets, **kwargs), 'dropna')
    common_index = intersection_of_index(features, targets)

    if len(features) <= 0:
        raise ValueError("not enough data!")

    return (
        features_and_labels.label_columns,
        features.loc[common_index],
        loc_if_not_none(targets, common_index)
    )


def extract(features_and_labels, df, extractor, *args, **kwargs):
    return features_and_labels(df, extractor, *args, **kwargs)


def extract_with_post_processor(list, postprocessor: Callable):
    def extractor(df, **kwargs):
        extraction = get_pandas_object(df, list, **kwargs)
        return get_pandas_object(extraction, postprocessor, **kwargs)

    return extractor




from typing import Tuple, Callable

import numpy as np
import pandas as pd


class Splitter(object):

    def __init__(self):
        pass

    def train_test_split(self, index: pd.Index) -> Tuple[pd.Index, pd.Index]:
        pass

    @staticmethod
    def infinity_sample_range():
        while True:
            yield

    @property
    def cross_validation(self) -> Tuple[int, Callable[[pd.Index, pd.Index], Tuple[np.ndarray, np.ndarray]]]:
        return None



import importlib.util as _ilu

from .auto_encoder_model import AutoEncoderModel
from .base_model import Model
from .keras_model import KerasModel
from .multi_model import MultiModel
from .pytoch_model import PytorchModel
from .scikit_learn_model import SkModel

if _ilu.find_spec("gym"):
    from .reinforcement_model import ReinforcementModel



from unittest import TestCase

from sklearn.neural_network import MLPRegressor, MLPClassifier

from pandas_ml_common import np, pd
from pandas_ml_utils_test.ml.data.model.test_abstract_model import TestAbstractModel
from pandas_ml_utils import SkModel, FeaturesAndLabels


class TestSkModel(TestAbstractModel, TestCase):

    def provide_classification_model(self, features_and_labels):
        model = SkModel(
            MLPClassifier(activation='logistic', max_iter=1000, hidden_layer_sizes=(3,), alpha=0.001, solver='lbfgs', random_state=42),
            features_and_labels,
        )

        return model

    def provide_regression_model(self, features_and_labels):
        model = SkModel(
            MLPRegressor(1, learning_rate_init=0.01, solver='sgd', activation='identity', momentum=0, max_iter=1500, n_iter_no_change=500, nesterovs_momentum=False, shuffle=False, validation_fraction=0.0, random_state=42),
            features_and_labels,
        )

        return model


import inspect
import logging
import math
from copy import deepcopy
from functools import partial
from typing import List, Callable, Union, Tuple

import numpy as np

from pandas_ml_common import Typing
from pandas_ml_common.utils import call_callable_dynamic_args
from pandas_ml_utils.ml.summary import Summary
from .base_model import Model
from ..data.splitting.sampeling import Sampler
from ..data.splitting.sampeling.extract_multi_model_label import ExtractMultiMultiModelSampler
from ..data.extraction import FeaturesAndLabels

_log = logging.getLogger(__name__)


class MultiModel(Model):

    def __init__(self,
                 basis_model: Model,
                 model_args: Union[int, List, Tuple],
                 model_index_variable: str = "i",
                 features_and_labels: FeaturesAndLabels = None,
                 summary_provider: Callable[[Typing.PatchedDataFrame], Summary] = Summary,
                 **kwargs):
        super().__init__(
            # we need to generate one label and sample weight for each of the sub models such that the global
            #  DataSampler returns one numpy array (which we then need to slice up again for each individual model).
            MultiModel._expand_features_and_labels(
                features_and_labels if features_and_labels is not None else basis_model.features_and_labels,
                MultiModel.models_iter(model_args),
                model_index_variable
            ),
            summary_provider,
            **kwargs
        )
        self.basis_model = basis_model
        self.sub_models: List[Model] = []
        self.model_index_variable = model_index_variable
        self.model_args = model_args
        self.nr_models = len(list(MultiModel.models_iter(model_args)))

    @staticmethod
    def _expand_features_and_labels(features_and_labels, model_args_iter, model_index_variable):
        # remove the variable which controls the avirous aspects of each of the multi model
        # we replace this information by the usage of a partial
        if model_index_variable in features_and_labels.kwargs:
            del features_and_labels.kwargs[model_index_variable]

        # early exit if nothing to do!
        if model_index_variable is None:
            return features_and_labels

        def wrap_partial(i, selector):
            if callable(selector):
                if model_index_variable in inspect.signature(selector).parameters.keys():
                    return partial(selector, **{model_index_variable: i})

            return selector

        labels = [wrap_partial(i, l) for i in model_args_iter for l in features_and_labels.labels if callable(l)]
        features_and_labels = features_and_labels.with_labels(labels)
        features_and_labels.set_label_columns(None, True)

        if features_and_labels.sample_weights is not None:
            weights = [wrap_partial(i, w) for i in model_args_iter for w in features_and_labels.sample_weights if callable(w)]
            features_and_labels = features_and_labels.with_sample_weights(weights)

        if features_and_labels.gross_loss is not None:
            gross_loss = [wrap_partial(i, gl) for i in model_args_iter for gl in features_and_labels.gross_loss if callable(gl)]
            features_and_labels = features_and_labels.with_gross_loss(gross_loss)

        return features_and_labels

    @staticmethod
    def models_iter(nr_of_models):
        return range(nr_of_models) if isinstance(nr_of_models, int) else nr_of_models

    def fit(self, sampler: Sampler, **kwargs) -> float:
        sub_model_losses = []
        nr_of_models = self.nr_models

        # copy sampler and add a new cross validator extracting the model
        for i_model, arg in enumerate(MultiModel.models_iter(self.model_args)):
            if "verbose" in kwargs and kwargs["verbose"]:
                print(f"fit model {i_model + 1} / {nr_of_models}")
            else:
                _log.info(f"fit model {i_model + 1} / {nr_of_models}")

            if "on_model" in kwargs:
                for callback in kwargs["on_model"]:
                    callback(i_model, {**kwargs, self.model_index_variable: arg})

            sm_loss = self.sub_models[i_model]\
                .fit(ExtractMultiMultiModelSampler(i_model, nr_of_models, sampler), **deepcopy(kwargs))
            sub_model_losses.append(sm_loss)

        self._history = [sm._history for sm in self.sub_models]
        return np.array(sub_model_losses).mean()

    def predict_sample(self, x: np.ndarray, **kwargs) -> np.ndarray:
        # call each sub_model.predict()
        predictions = [m.predict_sample(x) for m in self.sub_models]

        # and then concatenate the arrays back into one prediction
        return np.concatenate(predictions, 1) if predictions[0].ndim > 1 else np.stack(predictions, 1)

    def plot_loss(self, figsize=(8, 6), columns=None):
        # override the plot function and plot the loss per model and fold
        import matplotlib.pyplot as plt
        nr_of_models = self.nr_models

        columns = nr_of_models if columns is None else int(columns)
        rows = math.ceil(nr_of_models / columns)
        fig, axes = plt.subplots(rows, columns, figsize=(figsize if figsize else plt.rcParams.get('figure.figsize')))
        axes = axes.flatten()

        for m_nr, m_hist in enumerate(self._history):
            ax = axes[m_nr]

            for fold_nr, fold_loss in enumerate(m_hist):
                if len(fold_loss[0]) <= 1:
                    _log.warning("can not plot single epoch loss")

                p = ax.plot(fold_loss[0], '-', label=f'{m_nr}: {fold_nr}: loss')
                ax.plot(fold_loss[1], '--', color=p[-1].get_color(), label=f'{m_nr}: {fold_nr}: val loss')

        plt.legend(loc='upper right')
        return fig, axes

    def _slices(self, i, total_nr_of_columns=1) -> slice:
        nr_labels = total_nr_of_columns // self.nr_models
        return slice(i * nr_labels, (i + 1) * nr_labels)

    def __call__(self, *args, **kwargs):
        copy = deepcopy(self)
        nr_of_models = self.nr_models

        # get the labels and sample weights from the basis model
        l = self.features_and_labels.labels
        t = self.features_and_labels.targets
        w = self.features_and_labels.sample_weights
        gl = self.features_and_labels.gross_loss

        # this function is now reverse splitting up the generated labels
        # which is useful if one of the sum models gets used individually somewhere else
        def create_sub_model(i):
            # copy the FeaturesAndLabels object from the basis model
            fl = deepcopy(self.features_and_labels)

            # replace labels with one slice per model
            fl._labels = l[self._slices(i, len(l))]
            fl._targets = t[self._slices(i, len(t))] if t is not None else None
            fl._sample_weights = w[self._slices(i, len(w))] if w is not None else None
            fl._gross_loss = gl[self._slices(i, len(gl))] if gl is not None else None

            # create a new model from the basis model and replace the FeaturesAndLabels object
            sm = self.basis_model(*args, **kwargs)
            sm._features_and_labels = fl
            return sm

        # initialize all sub models and use the mutated FeaturesAndLabels objects we prepared for each model
        # Note! the data is actually engineered by the usage of the MultiModel's FeaturesAndLabels object and NOT
        # by the individual once. However this is needed to keep a FeaturesAndLabels and Models consistent.
        copy.sub_models = [create_sub_model(i) for i in range(nr_of_models)]
        return copy



from pandas_ml_common import Typing
from pandas_ml_utils import html
from pandas_ml_utils.constants import *
from pandas_ml_utils.ml.model.multi_model import MultiModel
from pandas_ml_utils.ml.summary import Summary


class MultiModelSummary(Summary):

    @staticmethod
    def filter_labels(df, model, nodel_nr):
        columns = [TARGET_COLUMN_NAME, PREDICTION_COLUMN_NAME, LABEL_COLUMN_NAME, GROSS_LOSS_COLUMN_NAME, SAMPLE_WEIGHTS_COLUMN_NAME, FEATURE_COLUMN_NAME]
        columns = [col for col in columns if col in df]
        filtered_columns = []

        for col in columns:
            if col != FEATURE_COLUMN_NAME:
                cols = df[[col]].columns.to_list()
                column_slice = model._slices(nodel_nr, len(cols))
                filtered_columns = [*filtered_columns, *cols[column_slice]]
            else:
                filtered_columns = [*filtered_columns, *df[[FEATURE_COLUMN_NAME]].columns.to_list()]

        return filtered_columns

    def __init__(self, df: Typing.PatchedDataFrame, model: MultiModel, summary_provider=None, **kwargs):
        super().__init__(df, model, **kwargs)
        self.summaries = \
            [(summary_provider or sm.summary_provider)(df[MultiModelSummary.filter_labels(df, model, i)], sm, **kwargs)
             for i, sm in enumerate(model.sub_models)]

    def _repr_html_(self):
        from mako.template import Template
        from mako.lookup import TemplateLookup

        template = Template(filename=html.SELF_TEMPLATE(__file__), lookup=TemplateLookup(directories=['/']))
        return template.render(fit=self, summaries=self.summaries)



from functools import partial
from unittest import TestCase

import numpy as np
import pandas as pd

from pandas_ml_utils import FeaturesAndLabels
from test.config import DF_TEST


class TestExtrationOfFeaturesAndLabels(TestCase):

    def test_repr(self):
        fnl = FeaturesAndLabels(
            features=[1, "A", lambda x: x * 2],
            labels=[1, "A", lambda x: x * 2],
        )

        # print(repr(fnl).replace('"', "\\\""))
        self.assertEqual(
            "FeaturesAndLabels(	[1, 'A', '            features=[1, \"A\", lambda x: x * 2],\\n'], 	[1, 'A', '            labels=[1, \"A\", lambda x: x * 2],\\n'], 	None, 	None, 	None, 	None, 	{})",
            repr(fnl)
        )

    def test__extract_label_comprehension(self):
        df = DF_TEST.copy()

        (features, _), labels, targets, weights, gross_loss = df._.extract(
            FeaturesAndLabels(
                features=["Close"],
                labels=[partial(lambda i, df: df["Close"].rename(f"Close_{i}"), i) for i in range(5, 10)]
            )
        )

        #print(labels.tail())
        self.assertListEqual(
            ["Close_5", "Close_6", "Close_7", "Close_8", "Close_9"],
            labels.columns.tolist()
        )

    def test__extract_inf(self):

        df = pd.DataFrame({"a": [1, 2, 4, np.nan, np.inf]})

        (features, _), labels, targets, weights, gross_loss = df._.extract(
            FeaturesAndLabels(
                features=["a"],
                labels=["a"]
            )
        )

        self.assertEqual(len(features), 3)

    def test_extraction(self):

        pass



import os
from copy import deepcopy
from typing import Callable, Tuple

import dill as pickle
import numpy as np

from pandas_ml_common import Typing
from pandas_ml_utils.ml.data.extraction import FeaturesAndLabels
from pandas_ml_utils.ml.data.splitting.sampeling import Sampler
from pandas_ml_utils.ml.summary import Summary


class Model(object):
    """
    Represents a statistical or ML model and holds the necessary information how to interpret the columns of a
    pandas *DataFrame* ( :class:`.FeaturesAndLabels` ). Currently available implementations are:
     * SkitModel - provide any skit learn classifier or regressor
     * KerasModel - provide a function returning a compiled keras model
     * MultiModel - provide a model which will copied (and fitted) for each provided target
    """

    @staticmethod
    def load(filename: str):
        """
        Loads a previously saved model from disk. By default `dill <https://pypi.org/project/dill/>`_ is used to
        serialize / deserialize a model.

        :param filename: filename of the serialized model inclusive file extension
        :return: returns a deserialized model
        """
        with open(filename, 'rb') as file:
            model = pickle.load(file)
            if isinstance(model, Model):
                return model
            else:
                raise ValueError("Deserialized pickle was not a Model!")

    def __init__(self,
                 features_and_labels: FeaturesAndLabels,
                 summary_provider: Callable[[Typing.PatchedDataFrame], Summary] = Summary,
                 **kwargs):
        """
        All implementations of `Model` need to pass two arguments to `super().__init()__`.

        :param features_and_labels: the :class:`.FeaturesAndLabels` object defining all the features,
                                    feature engineerings and labels
        :param summary_provider: a summary provider in the most simple case just holds a `pd.DataFrame` containing all
                                 the labels and all the predictions and optionally loss and target values. Since
                                 constructors as callables as well it is usually enough tho just pass the type i.e.
                                 `summary_provider=BinaryClassificationSummary`
        :param kwargs:
        """
        self._features_and_labels = features_and_labels
        self._summary_provider = summary_provider
        self._validation_indices = []
        self._history = []
        self.kwargs = kwargs

    @property
    def features_and_labels(self):
        return self._features_and_labels

    @property
    def summary_provider(self):
        return self._summary_provider

    @property
    def validation_indices(self):
        return self._validation_indices

    def __getitem__(self, item):
        """
        returns arguments which are stored in the kwargs filed. By providing a tuple, a default in case of missing
        key can be specified
        :param item: name of the item im the kwargs dict or tuple of name, default
        :return: item or default
        """
        if isinstance(item, tuple) and len(item) == 2:
            return self.kwargs[item[0]] if item[0] in self.kwargs else item[1]
        else:
            return self.kwargs[item] if item in self.kwargs else None

    def save(self, filename: str):
        """
        save model to disk
        :param filename: filename inclusive file extension
        :return: None
        """
        with open(filename, 'wb') as file:
            pickle.dump(self, file)

        print(f"saved model to: {os.path.abspath(filename)}")

    def fit(self, sampler: Sampler, **kwargs) -> float:
        """
        draws folds from the data generator as long as it yields new data and fits the model to one fold

        :param sampler: a data generating process class:`pandas_ml_utils.ml.data.splitting.sampeling.Sampler`
        :return: returns the average loss over oll folds
        """

        # sample: train[features, labels, target, weights], test[features, labels, target, weights]
        losses = [self.fit_fold(i, s[0][0], s[0][1], s[1][0], s[1][1], s[0][3], s[1][3], **kwargs)
                  for i, s in enumerate(sampler.sample())]

        self._history = losses

        # this loss is used for hyper parameter tuning so we take the average of the minimum loss of each fold
        return np.array([(fold_loss[0].min() if fold_loss[0].size > 0 else np.nan) for fold_loss in losses]).mean()

    def plot_loss(self, figsize=(8, 6), secondary_y=False, **kwargs):
        """
        plot a diagram of the training and validation losses per fold
        :return: figure and axis
        """

        import matplotlib.pyplot as plt
        fig, ax = plt.subplots(1, 1, figsize=(figsize if figsize else plt.rcParams.get('figure.figsize')))

        for fold_nr, fold_loss in enumerate(self._history):
            p = ax.plot(fold_loss[0], '-', label=f'{fold_nr}: loss')
            ax2 = ax.twinx() if secondary_y else ax
            ax2.plot(fold_loss[1], '--', color=p[-1].get_color(), label=f'{fold_nr}: val loss')

        plt.legend(loc='upper right')
        return fig, ax

    def predict(self, sampler: Sampler, **kwargs) -> np.ndarray:
        """
        predict as many samples as we can sample from the sampler

        :param sampler:
        :return:
        """
        # make shape (rows, samples, ...)
        return np.array([self.predict_sample(t[0]) for (t, _) in sampler.sample()]).swapaxes(0, 1)

    def fit_fold(self,
                 fold_nr: int,
                 x: np.ndarray, y: np.ndarray,
                 x_val: np.ndarray, y_val: np.ndarray,
                 sample_weight_train: np.ndarray, sample_weight_test: np.ndarray,
                 **kwargs) -> Tuple[np.ndarray, np.ndarray]:
        """
        function called to fit the model to one fold of the data generator (i.e. k-folds)
        :param fold_nr: number of fold in case of cross validation is used
        :param x: x
        :param y: y
        :param x_val: x validation
        :param y_val: y validation
        :param sample_weight_train: sample weights for loss penalisation (default np.ones)
        :param sample_weight_test: sample weights for loss penalisation (default np.ones)
        :return: loss of the fit
        """
        pass

    def predict_sample(self, x: np.ndarray, **kwargs) -> np.ndarray:
        """
        prediction of the model for each sample and target

        :param x: x
        :return: prediction of the model for each target
        """

        pass

    def __call__(self, *args, **kwargs):
        """
        returns a copy pf the model with eventually different configuration (kwargs). This is useful for hyper paramter
        tuning or for MultiModels

        :param args:
        :param kwargs: arguments which are eventually provided by hyperopt or by different targets
        :return:
        """
        if not kwargs:
            return deepcopy(self)
        else:
            raise ValueError(f"construction of model with new parameters is not supported\n{type(self)}: {kwargs}")




import logging
from typing import Tuple

import numpy as np
from sklearn.model_selection import train_test_split as sk_train_test_split

import pandas as pd

_log = logging.getLogger(__name__)


def train_test_split(index: pd.Index,
                     test_size: float = 0.4,
                     youngest_size: float = None,
                     seed: int = 42) -> Tuple[pd.Index, pd.Index]:

    # convert data frame index to numpy array
    index = index.values

    if test_size <= 0:
        train, test = index, index[:0]
    elif seed == 'youngest':
        i = int(len(index) - len(index) * test_size)
        train, test = index[:i], index[i:]
    else:
        random_sample_test_size = test_size if youngest_size is None else test_size * (1 - youngest_size)
        random_sample_train_index_size = int(len(index) - len(index) * (test_size - random_sample_test_size))

        if random_sample_train_index_size < len(index):
            _log.warning(f"keeping youngest {len(index) - random_sample_train_index_size} elements in test set")

            # cut the youngest data and use residual to randomize train/test data
            index_train, index_test = \
                sk_train_test_split(index[:random_sample_train_index_size],
                                 test_size=random_sample_test_size, random_state=seed)

            # then concatenate (add back) the youngest data to the random test data
            index_test = np.hstack([index_test, index[random_sample_train_index_size:]])  # index is 1D

            train, test = index_train, index_test
        else:
            train, test = sk_train_test_split(index, test_size=random_sample_test_size, random_state=seed)

    return pd.Index(train), pd.Index(test)




import math
from typing import Tuple, Callable, Any

import numpy as np
from matplotlib.axis import Axis
from matplotlib.figure import Figure
from mlxtend.evaluate import confusion_matrix
from sklearn.metrics import roc_curve, auc

from pandas_ml_common import Typing
from pandas_ml_common.utils.serialization_utils import plot_to_html_img
from pandas_ml_utils import html
from pandas_ml_utils.constants import *
from pandas_ml_utils.ml.model.base_model import Model

from pandas_ml_utils.ml.summary import Summary


class ReconstructionSummary(Summary):

    def __init__(self,
                 df: Typing.PatchedDataFrame,
                 model: Model,
                 reconstruction_plotter: Callable[[Axis, np.ndarray], Any] = lambda ax, x: ax.plot(x),
                 **kwargs):
        super().__init__(df, model, **kwargs)
        self.reconstruction_plotter = reconstruction_plotter

    def reconstruction_plot(self):
        import matplotlib.pyplot as plt

        _y = self.df[LABEL_COLUMN_NAME]._.values
        _y_hat = self.df[PREDICTION_COLUMN_NAME]._.values
        fig, ax = plt.subplots(2, 5, figsize=(20, 10))
        ax = ax.flatten()

        # random samples
        indices = np.random.randint(_y.shape[0], size=10)
        _y = _y[indices, :]
        _y_hat = _y_hat[indices, :]

        # print shape
        print(_y.shape, _y_hat.shape)

        for i in range(10):
            self.reconstruction_plotter(ax[i], _y[i])
            self.reconstruction_plotter(ax[i], _y_hat[i])

            ax[i].legend(["original", "reconstruct"])

        return fig

    def _repr_html_(self):
        from mako.template import Template
        from mako.lookup import TemplateLookup

        template = Template(filename=html.SELF_TEMPLATE(__file__), lookup=TemplateLookup(directories=['/']))
        return template.render(
            fit=self,
            recon_plot=plot_to_html_img(self.reconstruction_plot)
        )



from unittest import TestCase
import pandas as pd

from pandas_ml_utils.ml.data.splitting import NaiveSplitter


class TestNaiveSplitter(TestCase):

    def test_sampler(self):
        """given"""
        df = pd.DataFrame({"featureA": [1, 2, 3, 4, 5],
                           "labelA": [1, 2, 3, 4, 5]})

        """when"""
        train_ix, test_ix = NaiveSplitter().train_test_split(df.index)

        """then"""
        print(train_ix, test_ix)
        self.assertListEqual([0,1,2], train_ix.tolist())
        self.assertListEqual([3,4], test_ix.tolist())




from unittest import TestCase

import sys
import torch as t
import torch.nn as nn
from torch.optim import SGD

from pandas_ml_utils import pd, AutoEncoderModel, FeaturesAndLabels
from pandas_ml_utils.ml.model.pytoch_model import PytorchModel
from pandas_ml_utils_test.ml.data.model.test_abstract_model import TestAbstractModel


class RegressionModule(nn.Module):
    def __init__(self):
        super().__init__()
        self.regressor = nn.Sequential(
            nn.Linear(1, 1)
        )

    def forward(self, x):
        x = self.regressor(x)
        return x


class TestPytorchModel(TestAbstractModel, TestCase):

    def provide_classification_model(self, features_and_labels):
        t.manual_seed(42)

        def module_provider():
            class ClassificationModule(nn.Module):

                def __init__(self):
                    super().__init__()
                    self.classifier = nn.Sequential(
                        nn.Linear(2, 5),
                        nn.ReLU(),
                        nn.Linear(5, 1),
                        nn.Sigmoid()
                    )

                def forward(self, x):
                    x = self.classifier(x)
                    return x

            return ClassificationModule()

        model = PytorchModel(
            features_and_labels,
            module_provider,
            nn.MSELoss,
            lambda params: SGD(params, lr=0.03)
        )

        return model

    def provide_regression_model(self, features_and_labels):
        t.manual_seed(12)

        model = PytorchModel(
            features_and_labels,
            RegressionModule,
            nn.MSELoss,
            lambda params: SGD(params, lr=0.01, momentum=0.0)
        )

        return model

    def provide_auto_encoder_model(self, features_and_labels):
        t.manual_seed(12)

        def module_provider():
            class AutoEncoderModule(nn.Module):

                def __init__(self):
                    super().__init__()
                    self.encoder = nn.Sequential(
                        nn.Linear(2, 2),
                        nn.Tanh(),
                        nn.Linear(2, 1),
                        nn.Tanh(),
                    )

                    self.decoder = nn.Sequential(
                        nn.Linear(1, 2),
                        nn.Tanh(),
                        nn.Linear(2, 2),
                        nn.Tanh(),
                    )

                def forward(self, x):
                    x = self.encoder(x)
                    x = self.decoder(x)
                    return x

                def encode(self, x):
                    with t.no_grad():
                        return self.encoder(t.from_numpy(x).float()).numpy()

                def decode(self, x):
                    with t.no_grad():
                        return self.decoder(t.from_numpy(x).float()).numpy()

            return AutoEncoderModule()

        model = AutoEncoderModel(
            PytorchModel(
                features_and_labels,
                module_provider,
                nn.MSELoss,
                lambda params: SGD(params, lr=0.1, momentum=0.9)
            ),
            ["condensed"],
            lambda m: m.module.encode,
            lambda m: m.module.decode,
        )

        return model

    def test_callbacks(self):
        # a test with a early stopping callback and pass restore_best_weights=True as kwarg
        df = pd.DataFrame({
            "a": [1, 0, 1, 0, 1, 0, 1, 0, ],
            "b": [0, 1, 0, 1, 1, 0, 1, 0, ],
        })

        def module_provider():
            class ClassificationModule(nn.Module):

                def __init__(self):
                    super().__init__()
                    self.classifier = nn.Sequential(
                        nn.Linear(2, 5),
                        nn.ReLU(),
                        nn.Linear(5, 1),
                        nn.Sigmoid()
                    )

                def forward(self, x):
                    x = self.classifier(x)
                    return x

            return ClassificationModule()

        model = PytorchModel(
            FeaturesAndLabels(["a", "b"], ["b"]),
            module_provider,
            nn.MSELoss,
            lambda params: SGD(params, lr=0.1, momentum=0.9)
        )

        fit = df.model.fit(model, on_epoch=[PytorchModel.Callbacks.early_stopping(patience=3, tolerance=-100)], restore_best_weights=True)
        print(fit.model._history)
        self.assertEqual(4, len(fit.model._history[0][0]))



from typing import Tuple, Callable

import numpy as np
import pandas as pd
from sklearn.model_selection import TimeSeriesSplit

from pandas_ml_utils.ml.data.splitting.splitter import Splitter


class NaiveSplitter(Splitter):

    def __init__(self, test_size=0.3, nr_of_splits=None, epochs_per_split=1):
        super().__init__()
        self.test_size = test_size
        self._cross_validation = TimeSeriesSplit(nr_of_splits) if nr_of_splits is not None else None
        self._epochs_per_split = epochs_per_split

    def train_test_split(self, index: pd.Index) -> Tuple[pd.Index, pd.Index]:
        # we just split the sequence int past and recent data
        end_idx = int(len(index) * (1 - self.test_size))
        return index[0:end_idx], index[end_idx:]

    @property
    def cross_validation(self) -> Tuple[int, Callable[[pd.Index, pd.Index], Tuple[np.ndarray, np.ndarray]]]:
        if self._cross_validation is None:
            return None
        else:
            return (self._epochs_per_split, self._cross_validation.split)


# create an alias for convenience
TimeSeriesSplitter = NaiveSplitter


import logging
from copy import deepcopy
from typing import List, Callable

import numpy as np

from pandas_ml_common import Typing
from pandas_ml_utils.ml.summary import Summary
from .base_model import Model
from ..data.splitting.sampeling import Sampler

_log = logging.getLogger(__name__)


class AutoEncoderModel(Model):

    def __init__(self,
                 trainable_model: Model,
                 encoded_column_names: List[str],
                 encoder_provider: Callable[[Model], Callable[[np.ndarray], np.ndarray]],
                 decoder_provider: Callable[[Model], Callable[[np.ndarray], np.ndarray]],
                 summary_provider: Callable[[Typing.PatchedDataFrame], Summary] = Summary,
                 **kwargs):
        super().__init__(trainable_model.features_and_labels, summary_provider, **kwargs)
        self._features_and_labels = trainable_model.features_and_labels
        self.trainable_model = trainable_model
        self.encoded_column_names = encoded_column_names
        self.encoder_provider = encoder_provider
        self.decoder_provider = decoder_provider
        self.mode = 'train'

    @property
    def features_and_labels(self):
        return self._features_and_labels

    def as_trainable(self) -> Model:
        copy = self()
        copy.mode = 'train'
        copy._features_and_labels = deepcopy(self.trainable_model.features_and_labels)
        return copy

    def as_encoder(self) -> Model:
        fnl_copy = deepcopy(self.trainable_model.features_and_labels)
        fnl_copy.set_label_columns(self.encoded_column_names, True)

        copy = self()
        copy.mode = 'encode'
        copy._features_and_labels = fnl_copy
        return copy

    def as_decoder(self, decoder_features: List[Typing._Selector] = None) -> Model:
        if "bottleneck" in self.trainable_model.features_and_labels.kwargs and decoder_features is None:
            decoder_features = self.trainable_model.features_and_labels.kwargs["bottleneck"]

        fnl_copy = deepcopy(self.trainable_model.features_and_labels)
        fnl_copy._features = decoder_features

        copy = self()
        copy.mode = 'decode'
        copy._features_and_labels = fnl_copy
        return copy

    def fit(self, sampler: Sampler, **kwargs) -> float:
        loss = self.trainable_model.fit(sampler, **kwargs)
        self._history = self.trainable_model._history
        return loss

    def predict_sample(self, x: np.ndarray, **kwargs) -> np.ndarray:
        if x.ndim > 2 and x.shape[-1] == 1:
            x = x.reshape(x.shape[:-1])

        if self.mode == 'train':
            return self.trainable_model.predict_sample(x, **kwargs)
        elif self.mode == 'encode':
            return self.encoder_provider(self.trainable_model)(x, **kwargs)
        elif self.mode == 'decode':
            return self.decoder_provider(self.trainable_model)(x, **kwargs)

    def plot_loss(self, **kwargs):
        return self.trainable_model.plot_loss(**kwargs)

    def __call__(self, *args, **kwargs):
        copy = AutoEncoderModel(
            self.trainable_model(*args, **kwargs),
            self.encoded_column_names,
            self.encoder_provider,
            self.decoder_provider,
            self.summary_provider,
            **self.kwargs
        )

        copy.mode = self.mode
        return copy




class FitException(ValueError):

    def __init__(self, exception, model, *args, **kwargs):
        super().__init__(exception, *args, **kwargs)
        self._model = model

    @property
    def model(self):
        return self._model



from __future__ import annotations

import logging
import math
import sys
from copy import deepcopy
from typing import List, Callable, TYPE_CHECKING, Type, Dict, Tuple

import numpy as np

from pandas_ml_common import Typing
from pandas_ml_common.utils import call_callable_dynamic_args
from pandas_ml_common.utils.logging_utils import LogOnce
from pandas_ml_utils.ml.data.extraction import FeaturesAndLabels
from pandas_ml_utils.ml.summary import Summary
from .base_model import Model

_log = logging.getLogger(__name__)


class PytorchModel(Model):

    if TYPE_CHECKING:
        from torch.nn import Module, _Loss
        from torch.optim import Optimizer

    def __init__(self,
                 features_and_labels: FeaturesAndLabels,
                 module_provider: Type[Module],
                 criterion_provider: Type[_Loss],
                 optimizer_provider: Type[Optimizer],
                 summary_provider: Callable[[Typing.PatchedDataFrame], Summary] = Summary,
                 callbacks: Dict[str, List[Callable]] = {},
                 **kwargs):
        super().__init__(features_and_labels, summary_provider, **kwargs)
        self.module_provider = module_provider
        self.criterion_provider = criterion_provider
        self.optimizer_provider = optimizer_provider
        self.callbacks = callbacks
        self.module = None
        self.log_once = LogOnce().log

    def fit_fold(self,
                 fold_nr: int,
                 x: np.ndarray, y: np.ndarray,
                 x_val: np.ndarray, y_val: np.ndarray,
                 sample_weight: np.ndarray, sample_weight_val: np.ndarray,
                 **kwargs) -> Tuple[np.ndarray, np.ndarray]:
        # import specifics
        from torch.autograd import Variable
        import torch as t

        # TODO we should not re-initialize model, criterion and optimizer once we have it already
        #  TODO we might re-initialize the optimizer with a new fold with a changes learning rate?

        is_verbose = kwargs["verbose"] if "verbose" in kwargs else False
        on_epoch_callbacks = kwargs["on_epoch"] if "on_epoch" in kwargs else []
        restore_best_weights = kwargs["restore_best_weights"] if "restore_best_weights" in kwargs else False
        num_epochs = kwargs["epochs"] if "epochs" in kwargs else 100
        batch_size = kwargs["batch_size"] if "batch_size" in kwargs else 128
        use_cuda = kwargs["cuda"] if "cuda" in kwargs else False

        module = self.module
        criterion_provider = self.criterion_provider

        if use_cuda:
            criterion_provider = lambda: self.criterion_provider().cuda()
            module = module.cuda()

        module = module.train()
        criterion = criterion_provider()
        optimizer = self.optimizer_provider(module.parameters())
        best_model_wts = deepcopy(module.state_dict())
        best_loss = sys.float_info.max
        epoch_losses = []
        epoch_val_losses = []

        if hasattr(module, 'callback'):
            on_epoch_callbacks += [module.callback]

        if hasattr(criterion, 'callback'):
            on_epoch_callbacks += [criterion.callback]

        if is_verbose:
            print(f"fit fold {fold_nr} with {len(x)} samples in {math.ceil(len(x) / batch_size)} batches ... ")

        for epoch in range(num_epochs):
            for i in range(0, len(x), batch_size):
                nnx = Variable(t.from_numpy(x[i:i+batch_size])).float()
                nny = Variable(t.from_numpy(y[i:i+batch_size])).float()
                weights = Variable(t.from_numpy(sample_weight[i:i + batch_size])).float() \
                    if sample_weight is not None else t.ones(nny.shape[0])

                if use_cuda:
                    nnx, nny, weights = nnx.cuda(), nny.cuda(), weights.cuda()

                if nnx.shape[0] <= 1:
                    self.log_once("invalid_batch", _log.warning, "skip single element batch!")
                    continue

                # ===================forward=====================
                output = module(nnx)
                loss = self._calc_weighted_loss(criterion, output, nny, weights)

                # ===================backward====================
                optimizer.zero_grad()
                loss.backward()
                optimizer.step()

                if is_verbose > 1:
                    print(f"{epoch}:{i}\t{loss}\t")

            # ===================log========================
            # add validation loss history
            if y_val is not None and len(y_val) > 0:
                with t.no_grad():
                    nnx = t.from_numpy(x).float()
                    nny = t.from_numpy(y).float()
                    weights = t.from_numpy(sample_weight).float() \
                        if sample_weight is not None else t.ones(nny.shape[0])
                    nnx_val = t.from_numpy(x_val).float()
                    nny_val = t.from_numpy(y_val).float()
                    weights_val = t.from_numpy(sample_weight_val).float() \
                        if sample_weight_val is not None else t.ones(nny_val.shape[0])

                    if use_cuda:
                        nnx, nny = nnx.cuda(), nny.cuda()
                        nnx_val, nny_val = nnx_val.cuda(), nny_val.cuda()
                        weights, weights_val = weights.cuda(), weights_val.cuda()

                    y_hat = module(nnx)
                    loss = self._calc_weighted_loss(criterion_provider(), y_hat, nny, weights).item()
                    epoch_losses.append(loss)

                    y_hat_val = module(nnx_val)
                    val_loss = self._calc_weighted_loss(criterion_provider(), y_hat_val, nny_val, weights_val).item()
                    epoch_val_losses.append(val_loss)

                    if val_loss < best_loss:
                        best_loss = val_loss
                        best_model_wts = deepcopy(module.state_dict())

            if is_verbose:
                print(f"{epoch}\t{loss}\t{val_loss}")

            # invoke on epoch end callbacks
            try:
                for callback in on_epoch_callbacks:
                    call_callable_dynamic_args(callback,
                                               fold=fold_nr,
                                               epoch=epoch,
                                               x=x, y=y, x_val=x_val, y_val=y_val,
                                               y_hat=y_hat, y_hat_val=y_hat_val,
                                               loss=loss, val_loss=val_loss, best_loss=best_loss)
            except StopIteration:
                break

        if restore_best_weights:
            module.load_state_dict(best_model_wts)

        return np.array(epoch_losses), np.array(epoch_val_losses)

    def _calc_weighted_loss(self, criterion, y_hat, y, weights):
        loss = criterion(y_hat, y)

        if loss.ndim > 0:
            if loss.ndim == weights.ndim:
                loss = (loss * weights).mean()
            else:
                self.log_once("loss.ndim!=weights.ndim", _log.warning,
                              f"sample weight has different dimensions {loss.shape}, {weights.shape}")

                if weights.ndim > loss.ndim and weights.shape[-1] == 1:
                    loss = self._calc_weighted_loss(criterion, y_hat, y, weights.squeeze())
                else:
                    loss = (loss * weights.repeat(1, *loss.shape[1:])).mean()

        return loss

    def predict_sample(self, x: np.ndarray, **kwargs) -> np.ndarray:
        # import specifics
        import torch as t

        with t.no_grad():
            module = self.module.cpu().eval()
            res = module(t.from_numpy(x).float())
            return res if isinstance(res, np.ndarray) else res.numpy()

    def __getstate__(self):
        # Copy the object's state from self.__dict__ which contains all our instance attributes.
        # Always use the dict.copy() method to avoid modifying the original state.
        state = self.__dict__.copy()

        # remove un-pickleable fields
        del state['module']

        # add torch serialisation
        state['module_state_dict'] = self.module.state_dict() if self.module is not None else None

        # return altered state
        return state

    def __setstate__(self, state):
        # use torch.save(model.state_dict(), './sim_autoencoder.pth')
        # first remove the special state
        module_state_dict = state['module_state_dict']
        del state['module_state_dict']

        # Restore instance attributes
        self.__dict__.update(state)
        self.module = self.module_provider()

        # restore special state dict
        if module_state_dict is not None:
            self.module.load_state_dict(module_state_dict)

    def __call__(self, *args, **kwargs):
        pytorch_model = PytorchModel(
            self.features_and_labels,
            self.module_provider,
            self.criterion_provider,
            self.optimizer_provider,
            self.summary_provider,
            deepcopy(self.callbacks)
        )

        pytorch_model.module = self.module_provider()

        # copy weights of existing models
        if self.module is not None:
            pytorch_model.module.load_state_dict(deepcopy(self.module.state_dict()))

        return pytorch_model

    # Add some useful callbacks directly to the pytorch model
    class Callbacks(object):

        @staticmethod
        def print_loss(mod=10):
            def printer(fold, epoch, loss, val_loss):
                if epoch % mod == 0:
                    print(f"{fold}: {epoch}: {loss}\t {val_loss}")

            return printer

        class early_stopping(object):

            def __init__(self, patience=1, tolerance=0.001):
                self.patience = patience
                self.tolerance = tolerance
                self.last_loss = sys.float_info.max
                self.counter = 0

            def __call__(self, val_loss, **kwargs):
                if (val_loss - self.tolerance) < self.last_loss:
                    self.last_loss = val_loss
                    self.counter = 0
                else:
                    self.counter += 1
                    if self.counter >= self.patience:
                        print(f"early stopping {self.counter}, {val_loss} > {self.last_loss}")
                        raise StopIteration("early stopping")

            def __copy__(self):
                return type(self)(self.patience, self.tolerance)



import pandas_ml_utils.html as html
from pandas_ml_common import Typing


class Summary(object):
    """
    Summary objects a used to visually present the results of a `df.fit` fitted model or a `df.backtest`
    All implementations of `Summary` need to:
     * pass a `pd.DataFrame` to `super().__init()__`
     * implement `_repr_html_()`
    """

    def __init__(self, df: Typing.PatchedDataFrame, model: 'Model', **kwargs):
        self._df = df
        self.model = model
        self.kwargs = kwargs

    @property
    def df(self):
        return self._df

    def _repr_html_(self):
        from mako.template import Template
        from mako.lookup import TemplateLookup

        template = Template(filename=html.SELF_TEMPLATE(__file__), lookup=TemplateLookup(directories=['/']))
        return template.render(summary=self)




import inspect
from copy import deepcopy
from typing import List, Callable, Tuple, Union, Any, TypeVar

import pandas as pd

from pandas_ml_common.utils.callable_utils import call_callable_dynamic_args
from pandas_ml_utils.ml.data.extraction.features_and_labels_extractor import extract_feature_labels_weights
from pandas_ml_common import Typing

T = TypeVar('T', str, List, Callable[[Any], Union[pd.DataFrame, pd.Series]])


# This class should be able to be pickled and unpickled without risk of change between versions
# This means business logic need to be kept outside of this class!
class FeaturesAndLabels(object):
    """
    *FeaturesAndLabels* is the main object used to hold the context of your problem. Here you define which columns
    of your `DataFrame` is a feature, a label or a target. This class also provides some functionality to generate
    autoregressive features. By default lagging features results in an RNN shaped 3D array (in the format of keras
    RNN layers input format).
    """

    def __init__(self,
                 features: Union[Typing._Selector,Tuple[Typing._Selector]],
                 labels: Typing._Selector = [],
                 sample_weights: Typing._Selector = None,
                 gross_loss: Typing._Selector = None,
                 targets: Typing._Selector = None,
                 label_type = None,
                 **kwargs):
        """
        :param features: a list of column names which are used as features for your model
        :param labels: as list of column names which are uses as labels for your model. you can specify one ore more
            named targets for a set of labels by providing a dict. This is useful if you want to train a
            :class:`.MultiModel` or if you want to provide extra information about the label. i.e. you
            want to classify whether a stock price is below or above average and you want to provide what
            the average was. It is also possible to provide a Callable[[df, ...magic], labels] which returns
            the expected .data structure.
        :param sample_weights: sample weights get passed to the model.fit function. In keras for example this can be
             used for imbalanced classes
        :param gross_loss: expects a callable[[df, target, ...magic], df] which receives the source .data frame and a
            target (or None) and should return a series or .data frame. Let's say you want to classify whether a
            printer is jamming the next page or not. Halting and servicing the printer costs 5'000 while a jam costs
            15'000. Your target will be 0 or empty but your gross loss will be -5000 for all your type II errors and
            -15'000 for all your type I errors in case of miss-classification. Another example would be if you want to
            classify whether a stock price is above (buy) the current price or not (do nothing). Your target is the
            today's price and your loss is tomorrows price minus today's price.
        :param targets: expects a callable[[df, targets, ...magic], df] which receives the source .data frame and a
            target (or None) and should return a series or .data frame. In case of multiple targets the series names
            need to be unique!
        :param min_required_samples: in case you only want to do one prediction and have some feature engineering
            you might need some minimum amount of samples to engineer your features. This can be None but it increases
            performance if you provide it.
        :param label_type: cast label to its type after all nans got dropped
        :param kwargs: maybe you want to pass some extra parameters to a callable you have provided
        """

        self._features = features
        self._labels = labels
        self._label_type = label_type
        self._sample_weights = sample_weights
        self._targets = targets
        self._gross_loss = gross_loss
        self._kwargs = kwargs

        # set after fit
        self._min_required_samples = None
        self._label_columns = None

    def set_min_required_samples(self, min_required: int):
        self._min_required_samples = min_required

        return self

    def set_label_columns(self, column_names, over_write=False) -> 'FeaturesAndLabels':
        if not self._label_columns or over_write:
            self._label_columns = column_names

        return self

    @property
    def features(self):
        return self._features

    @property
    def labels(self):
        return self._labels

    @property
    def label_type(self):
        return self._label_type

    @property
    def sample_weights(self):
        return self._sample_weights

    @property
    def targets(self):
        return self._targets

    @property
    def gross_loss(self):
        return self._gross_loss

    @property
    def min_required_samples(self) -> Union[int, Callable[[Any], int]]:
        return self._min_required_samples

    @property
    def label_columns(self):
        return self._label_columns

    @property
    def kwargs(self):
        return self._kwargs

    def with_labels(self, labels: Union[str, List[T], Callable[[Any], Union[pd.DataFrame, pd.Series]]]):
        copy = deepcopy(self)
        copy._labels = labels
        return copy

    def with_sample_weights(self, sample_weights: Union[str, List[T], Callable[[Any], Union[pd.DataFrame, pd.Series]]]):
        copy = deepcopy(self)
        copy._sample_weights = sample_weights
        return copy

    def with_gross_loss(self, gross_loss: Union[str, List[T], Callable[[Any], Union[pd.DataFrame, pd.Series]]]):
        copy = deepcopy(self)
        copy._gross_loss = gross_loss
        return copy

    def with_kwargs(self, **kwargs):
        copy = deepcopy(self)
        copy.kwargs = {**self.kwargs, **kwargs}
        return copy

    def __call__(self,
                 df: pd.DataFrame,
                 extractor: callable = extract_feature_labels_weights,
                 *args,
                 **kwargs) -> Union[pd.DataFrame, Tuple[pd.DataFrame, pd.DataFrame, pd.Series]]:
        # Basic concept:
        #  we call an extractor(df, **{**self.kwargs, **kwargs})
        #  this extractor uses 'get_pandas_object' which itself can handle lambdas with dependecies
        #  injected from available kwargs
        return call_callable_dynamic_args(extractor, df, self, **{**self.kwargs, **kwargs})

    def __hash__(self):
        return hash(self.__id__())

    def __id__(self):
        return self.__repr__()

    def __str__(self):
        return self.__repr__()

    def __eq__(self, other):
        return self.__id__() == other.__id__() if isinstance(other, FeaturesAndLabels) else False

    def __repr__(self):
        def source(params):
            if params is None:
                return None
            else:
                def insp(p):
                    try:
                        return inspect.getsource(p)
                    except Exception as e:
                        return p

                return [insp(p) if callable(p) else p if isinstance(p, (bool, int, float, str)) else repr(p) for p in params]

        return f'FeaturesAndLabels(' \
               f'\t{source(self.features)}, ' \
               f'\t{source(self.labels)}, ' \
               f'\t{source(self.sample_weights)}, ' \
               f'\t{source(self.gross_loss)}, ' \
               f'\t{source(self.targets)}, ' \
               f'\t{self.label_type}, ' \
               f'\t{self.kwargs}' \
               f')'




from __future__ import annotations

import logging
from copy import deepcopy
from typing import Callable, Tuple

import numpy as np
from sklearn.exceptions import ConvergenceWarning
from sklearn.linear_model import LogisticRegression

from pandas_ml_common import Typing
from pandas_ml_utils.ml.data.extraction import FeaturesAndLabels
from pandas_ml_utils.ml.summary import Summary
from .base_model import Model

_log = logging.getLogger(__name__)
ConvergenceWarning('ignore')


class SkModel(Model):

    def __init__(self,
                 skit_model,
                 features_and_labels: FeaturesAndLabels,
                 summary_provider: Callable[[Typing.PatchedDataFrame], Summary] = Summary,
                 **kwargs):
        super().__init__(features_and_labels, summary_provider, **kwargs)
        self.skit_model = skit_model
        self.label_shape = None

    def fit_fold(self,
                 fold_nr: int,
                 x: np.ndarray, y: np.ndarray,
                 x_val: np.ndarray, y_val: np.ndarray,
                 sample_weight_train: np.ndarray, sample_weight_test: np.ndarray,
                 **kwargs) -> Tuple[np.ndarray, np.ndarray]:
        # shape correction if needed
        x = SkModel.reshape_rnn_as_ar(x)
        y = y.reshape((len(x), -1)) if y.ndim > 1 and y.shape[1] == 1 else y
        y = y.reshape(len(x)) if y.ndim == 2 and y.shape[1] == 1 else y
        self.label_shape = y.shape

        # remember fitted model
        self.skit_model = self.skit_model.fit(x, y)

        if getattr(self.skit_model, 'loss_', None):
            loss_curve = getattr(self.skit_model, 'loss_curve_', [])
            return np.array(loss_curve), np.array([])
        else:
            prediction = self.predict_sample(x)
            if isinstance(self.skit_model, LogisticRegression)\
            or type(self.skit_model).__name__.endswith("Classifier")\
            or type(self.skit_model).__name__.endswith("SVC"):
                from sklearn.metrics import log_loss
                try:
                    return np.array(log_loss(prediction > 0.5, y).mean()), np.array([])
                except ValueError as e:
                    if "contains only one label" in str(e):
                        return np.array([-100]), np.array([])
                    else:
                        raise e
            else:
                from sklearn.metrics import mean_squared_error
                return np.array(mean_squared_error(prediction, y).mean()), np.array([])

    def predict_sample(self, x: np.ndarray, **kwargs) -> np.ndarray:
        if callable(getattr(self.skit_model, 'predict_proba', None)):
            y_hat = self.skit_model.predict_proba(SkModel.reshape_rnn_as_ar(x))
            binary_classifier = len(self.label_shape) == 1 or self.label_shape[1] == 1
            return y_hat[:, 1] if binary_classifier else y_hat.reshape(-1, *self.label_shape[1:])
        else:
            return self.skit_model.predict(SkModel.reshape_rnn_as_ar(x))

    def __str__(self):
        return f'{__name__}({repr(self.skit_model)}, {self.features_and_labels})'

    def __call__(self, *args, **kwargs):
        if not kwargs:
            return deepcopy(self)
        else:
            new_model = SkModel(type(self.skit_model)(**kwargs), self.features_and_labels, self.summary_provider)
            new_model.kwargs = deepcopy(self.kwargs)
            return new_model

    @staticmethod
    def reshape_rnn_as_ar(arr3d):
        if arr3d.ndim < 3:
            print("Data was not in RNN shape")
            return arr3d
        else:
            return arr3d.reshape(arr3d.shape[0], np.array(arr3d.shape[1:]).prod())



from unittest import TestCase

import torch.nn as nn
from torch.optim import Adam

from pandas_ml_utils import PytorchModel, FeaturesAndLabels
from pandas_ml_utils import pd, np
from pandas_ml_utils.ml.data.splitting import NaiveSplitter
from pandas_ml_utils.pytorch.loss import MultiObjectiveLoss


class TestLoss(TestCase):

    def test_multi_objective_loss(self):
        df = pd.DataFrame(np.array([
            # train
            [0, 0, 0],
            [0, 1, 1],
            [1, 0, 1],
            [1, 1, 1],
            # test
            [0, 0, 0],
            [0, 1, 1],
            [1, 0, 1],
            [1, 1, 1],
        ]), columns=["f1", "f2", "l"])

        class XorModule(nn.Module):

            def __init__(self):
                super().__init__()
                self.x1 = nn.Linear(2, 1)
                self.s1 = nn.Sigmoid()
                self.x2 = nn.Linear(2, 1)
                self.s2 = nn.Sigmoid()
                self.s = nn.Softmax()

            def forward(self, x):
                if self.training:
                    return self.s1(self.x1(x)), self.s2(self.x2(x))
                else:
                    return self.s1(self.x1(x))

        fit = df.model.fit(
            PytorchModel(
                FeaturesAndLabels(["f1", "f2"], ["l"]),
                XorModule,
                lambda: MultiObjectiveLoss((1, nn.MSELoss(reduction='none')),
                                           (1, nn.L1Loss(reduction='none')),
                                           on_epoch=lambda criterion, epoch: criterion.update_weights((0, 1.1))),
                Adam
            ),
            NaiveSplitter(0.5)
        )

        print(fit.test_summary.df)





from .splitter import Splitter
from .random_splits import RandomSplits
from .random_sequences import RandomSequences
from .naive import NaiveSplitter, TimeSeriesSplitter
from .dummy import DummySplitter


from typing import Tuple, Callable

import torch as t
import torch.nn as nn
from torch.nn.modules.loss import _Loss


class MultiObjectiveLoss(nn.Module):

    def __init__(self,
                 *criterion: Tuple[int, _Loss],
                 on_epoch: Callable[['MultiObjectiveLoss', t.Tensor], t.Tensor] = None,
                 reduction='mean',
                 max_weight=1,
                 min_weight=0,
                 ):
        super().__init__()
        self.weights = [c[0] for c in criterion]
        self.losses = [c[1] for c in criterion]
        self.on_epoch = on_epoch
        self.reduction = reduction
        self.max_weight = max_weight
        self.min_weight = min_weight

    def forward(self, y_hat, y):
        # calculate individual losses
        losses = [self.losses[i](y_hat[i], y) for i in range(len(self.losses))]

        # eventually reduce dimension to (batch, ) and apply weight on each sample of the bach
        losses = [(loss.sum(dim=1) if loss.ndim > 1 else loss) * self.weights[i] for i, loss in enumerate(losses)]

        # stack losses and return a sum per batch
        if self.reduction == 'sum':
            s = losses[0].sum()
            for i in range(1, len(losses)):
                s += losses[i].sum()

            return s
        elif self.reduction == 'mean':
            s = losses[0].mean()
            for i in range(1, len(losses)):
                s += losses[i].mean()

            return s
        else:
            return t.stack(losses, dim=1).sum(dim=1)

    def update_weights(self, *factors: Tuple[int, float]):
        for factor in factors:
            self.weights[factor[0]] = max(min(self.weights[factor[0]] * factor[1], self.max_weight), self.min_weight)

    def callback(self, epoch):
        if self.on_epoch is not None:
            self.on_epoch(self, epoch)


class CrossEntropyLoss(nn.Module):

    def __init__(self):
        super().__init__()
        self.cel = nn.CrossEntropyLoss(reduction='none')

    def forward(self, y_pred, y_true):
        y_true = y_true.view(y_pred.shape)

        if y_true.ndim > 2:
            if y_true.shape[1] > 1:
                return t.stack([self.forward(y_pred[:, i], y_true[:, i]) for i in range(y_true.shape[1])], dim=1)
            else:
                return self.cel(y_pred[:, 0], y_true[:, 0].argmax(dim=-1))
        else:
            return self.cel(y_pred, y_true.argmax(dim=-1))



import logging
from typing import Tuple, Callable, List, Generator

import numpy as np

from pandas_ml_common import Typing
from pandas_ml_common.utils import loc_if_not_none
from pandas_ml_utils.ml.data.splitting import Splitter

_log = logging.getLogger(__name__)


class Sampler(object):

    def __init__(self,
                 train: List[Typing.PatchedDataFrame],
                 test: List[Typing.PatchedDataFrame],
                 cross_validation: Tuple[int, Callable] = None):
        self.train = train
        self.test = test

        # add a default fold epoch of 1
        self._cross_validation = (1, cross_validation) if callable(cross_validation) else cross_validation

    def __getitem__(self, item) -> Tuple[Typing.PatchedDataFrame, Typing.PatchedDataFrame]:
        return self.train[item], self.test[item]

    @property
    def nr_of_source_events(self) -> Tuple[int, int]:
        return (
            len(self.train[0]) if self.train is not None and len(self.train) > 0 else -1,
            len(self.test[0]) if self.test is not None and len(self.test) > 0 else -1
        )

    def training(self) -> Tuple['Sampler', Typing.PdIndex]:
        return Sampler(self.train, []), self.train[0].index

    def validation(self) -> Tuple['Sampler', Typing.PdIndex]:
        return Sampler(self.test, []), self.test[0].index

    def sample(self) -> Generator[Tuple[List[np.ndarray], List[np.ndarray]], None, None]:
        train = [t._.values if t is not None else None for t in self.train]
        test = [t._.values if t is not None else None for t in self.test]
        cv = self._cross_validation

        # loop through folds and yield data until done then raise StopIteration
        if cv is not None and isinstance(cv, Tuple) and callable(cv[1]):
            for fold_epoch in range(cv[0]):
                # cross validation, make sure we re-shuffle every fold_epoch
                for f, (train_idx, test_idx) in enumerate(cv[1](train[0], train[1])):
                    _log.info(f'fit fold {f}')
                    yield ([t[train_idx] if t is not None else None for t in train],
                           ([t[test_idx] if t is not None else None for t in train] if test_idx is not None else []))
        else:
            # fit without cross validation
            yield train, test


class DataGenerator(object):
    # the idea is to pass a yielding data generator to the "fit" method instad of x/y values
    # the cross validation loop goes as default implementation into to Model class.
    # this way each model can implement cross validation on their own and we can use the data generator for the gym

    def __init__(self, splitter: Splitter, *frames: Typing.PatchedDataFrame):
        self.splitter = splitter
        self.frames = frames

    def train_test_sampler(self) -> Sampler:
        train_idx, test_idx = self.splitter.train_test_split(self.frames[0].index)
        train = [loc_if_not_none(frame, train_idx) for frame in self.frames]
        test = [loc_if_not_none(frame, test_idx) for frame in self.frames]
        return Sampler(train, test, self.splitter.cross_validation)

    def complete_samples(self) -> Sampler:
        return Sampler(self.frames, [], self.splitter.cross_validation)


import os

import numpy as np

from pandas_ml_common import pd


DF_NOTES = pd.read_csv(os.path.join(os.path.dirname(os.path.abspath(__file__)), ".data", "banknote_authentication.csv"))
DF_DEBUG = pd.DataFrame({"Close": np.random.random(10)})



import os

TEMLATE_DIR = os.path.dirname(os.path.abspath(__file__))
FIT_TEMPLATE = os.path.join(TEMLATE_DIR, "fit.py.html")


def SELF_TEMPLATE(file):
    return os.path.join(TEMLATE_DIR, f"{os.path.basename(file)}.html")




class FeatureScoreReport(object):

    def __init__(self, cv_fig, lags_fig, scores):
        self.cv_fig = cv_fig
        self.lags_fig = lags_fig
        self.scores = scores


class FeatureSelectionReport(object):

    def __init__(self):
        # return a report with a nice _html_repr_ with plots
        # and return an automatically generated FeaturesAndLabels object with the best selected features for the given label
        pass



import numpy as np
import pandas as pd
from unittest import TestCase

from pandas_ml_utils.ml.data.splitting import RandomSequences, DummySplitter
from pandas_ml_utils.ml.data.splitting.sampeling import DataGenerator, Sampler
from pandas_ml_utils.ml.data.splitting.sampeling.extract_multi_model_label import ExtractMultiMultiModelSampler


class TestDataGenerator(TestCase):

    def test_generate_single_sample(self):
        data = pd.DataFrame({"a": np.arange(10)})
        test = (data, data, None)
        train = (data, data, None)
        gen = Sampler(train, test)

        samples = [s for s in gen.sample()]

        self.assertEqual(1, len(samples))

    def test_generate_multiple_sample(self):
        data = pd.DataFrame({"a": np.arange(10)})
        test = (data, data, None)
        train = (data, data, None)
        gen = Sampler(train, test, cross_validation=RandomSequences(0, 0.5, 12).cross_validation)

        samples = [s for s in gen.sample()]

        self.assertEqual(12, len(samples))

    def test_dummy_generator(self):
        features = pd.DataFrame({"a": np.arange(10)})
        targets = pd.DataFrame({"a": np.arange(10)})

        one = DataGenerator(DummySplitter(1), features, targets).complete_samples()
        two = DataGenerator(DummySplitter(2), features, targets).complete_samples()

        sampled1 = [s for s in one.sample()]
        sampled2 = [s for s in two.sample()]

        self.assertEqual(1, len(sampled1))
        self.assertEqual(2, len(sampled2))

    def test_wrapping_cross_validation(self):
        features = pd.DataFrame({"a": np.arange(10)})
        labels = pd.DataFrame({
            "a": np.arange(10),
            "b": np.arange(10) + 10,
            "c": np.arange(10) + 20,
            "d": np.arange(10) + 30,
        })

        s = DataGenerator(DummySplitter(2), features, labels, None, None, None).train_test_sampler()
        s1 = ExtractMultiMultiModelSampler(0, 2, s)
        s2 = ExtractMultiMultiModelSampler(1, 2, s)

        train, test = next(s.sample())
        y = train[1]
        self.assertEqual(4, y.shape[1])
        self.assertEqual(780, y.sum())

        train, test = next(s1.sample())
        y = train[1]
        self.assertEqual(2, y.shape[1])
        self.assertEqual(190, y.sum())

        train, test = next(s2.sample())
        y = train[1]
        self.assertEqual(2, y.shape[1])
        self.assertEqual(590, y.sum())

        for s in s1.sample():
            print("----")
            print(s[0][0])

        self.assertEqual(20, sum([len(s[0][0]) for s in s1.sample()]))
        self.assertEqual(20, sum([len(s[0][0]) for s in s2.sample()]))



# TODO move to ml-common.plot as pair_plot(df, label_column) and use get_pandas_object
#  once the FIXMEs are fixed


def plot_features(joined_features_andLabels_df, label_column):
    import seaborn as sns

    # fixme if labels are contonious, we need to bin them
    # fixme if one hot encoded label column use np.argmax
    return sns.pairplot(joined_features_andLabels_df, hue=label_column)


from unittest import TestCase

import numpy as np
from sklearn.neural_network import MLPClassifier
from sklearn.model_selection import KFold

from pandas_ml_utils import FeaturesAndLabels, SkModel
from pandas_ml_utils.constants import *
from pandas_ml_utils.ml.data.splitting import NaiveSplitter, RandomSplits
from pandas_ml_utils_test.config import DF_NOTES


class TestModel(TestCase):

    def test_simple_classification_model(self):
        df = DF_NOTES.copy()

        fit = df.model.fit(
            SkModel(
                MLPClassifier(activation='tanh', hidden_layer_sizes=(20, 12), random_state=42, max_iter=2),
                FeaturesAndLabels(
                    features=["variance", "skewness", "kurtosis", "entropy"],
                    labels=["authentic"],
                    label_type=bool
                )
            ),
            NaiveSplitter()
        )

        print(fit)
        html = fit._repr_html_()

        prediction = df.model.predict(fit.model)
        print(prediction)
        self.assertGreaterEqual(prediction[PREDICTION_COLUMN_NAME].iloc[-1].values, 0.68)

        backtest = df.model.backtest(fit.model)
        self.assertIn(FEATURE_COLUMN_NAME, backtest.df)
        self.assertIn(LABEL_COLUMN_NAME, backtest.df)
        np.testing.assert_array_almost_equal(prediction[PREDICTION_COLUMN_NAME].iloc[-1].values,
                                             backtest.df[PREDICTION_COLUMN_NAME].iloc[-1].values)

        # test multiple samples
        samples = df.model.predict(fit.model, samples=2)
        self.assertIsInstance(samples[PREDICTION_COLUMN_NAME].iloc[-1, 0], list)
        self.assertEqual(2, len(samples[PREDICTION_COLUMN_NAME].iloc[-1, 0]))

    def test_simple_classification_cross_validation(self):
        df = DF_NOTES.copy()

        fit = df.model.fit(
            SkModel(
                MLPClassifier(activation='tanh', hidden_layer_sizes=(20, 12), random_state=42, max_iter=2),
                FeaturesAndLabels(
                    features=["variance", "skewness", "kurtosis", "entropy"],
                    labels=["authentic"],
                    label_type=bool
                )
            ),
            RandomSplits(cross_validation=KFold(3, random_state=42).split)
        )

        print(fit)
        html = fit._repr_html_()

        prediction = df.model.predict(fit.model)
        print(prediction)
        self.assertGreaterEqual(prediction[PREDICTION_COLUMN_NAME].iloc[-1].values, 0.7)

    def test_simple_classification_model_with_all_options(self):
        df = DF_NOTES.copy()

        fit = df.model.fit(
            SkModel(
                MLPClassifier(activation='tanh', hidden_layer_sizes=(20, 12), random_state=42, max_iter=2),
                FeaturesAndLabels(
                    features=["variance", "skewness", "kurtosis", "entropy"],
                    sample_weights=["variance"],
                    gross_loss=["kurtosis"],
                    targets=["entropy"],
                    labels=["authentic"],
                    label_type=bool
                )
            ),
            NaiveSplitter()
        )

        # should not thro an error
        html = fit._repr_html_()

        # fit resulting columns
        print(fit.test_summary.df)
        self.assertEqual(int(len(df) * 0.7), len(fit.training_summary.df))
        self.assertEqual(len(df) - int(len(df) * 0.7), len(fit.test_summary.df))
        # FIXME expect gross_loss and weights as well
        self.assertIn(FEATURE_COLUMN_NAME, fit.training_summary.df)
        self.assertIn(LABEL_COLUMN_NAME, fit.training_summary.df)
        self.assertIn(TARGET_COLUMN_NAME, fit.training_summary.df)

        self.assertIn(FEATURE_COLUMN_NAME, fit.test_summary.df)
        self.assertIn(LABEL_COLUMN_NAME, fit.test_summary.df)
        self.assertIn(TARGET_COLUMN_NAME, fit.test_summary.df)

        # prediction resulting columns
        prediction = df.model.predict(fit.model)
        print(prediction)
        self.assertIn(FEATURE_COLUMN_NAME, prediction)
        self.assertIn(TARGET_COLUMN_NAME, prediction)

        # backtest resulting columns
        # FIXME expect gross_loss and weights as well
        backtest = df.model.backtest(fit.model)
        print(backtest.df)
        self.assertEqual(len(df), len(backtest.df))
        self.assertIn(FEATURE_COLUMN_NAME, backtest.df)
        self.assertIn(LABEL_COLUMN_NAME, backtest.df)
        self.assertIn(TARGET_COLUMN_NAME, backtest.df)


from typing import Callable, Tuple, Dict, Union, Iterable

from pandas_ml_common import Typing
from pandas_ml_common.utils import has_indexed_columns, merge_kwargs
from pandas_ml_utils.ml.data.analysis import feature_selection
from pandas_ml_utils.ml.data.analysis.plot_features import plot_features
from pandas_ml_utils.ml.data.extraction import extract_feature_labels_weights
from pandas_ml_utils.ml.data.extraction.features_and_labels_definition import FeaturesAndLabels
from pandas_ml_utils.ml.data.splitting import Splitter, RandomSplits
from pandas_ml_utils.ml.fitting import fit, backtest, predict, Fit
from pandas_ml_utils.ml.model import Model as MlModel
from pandas_ml_utils.ml.summary import Summary


class Model(object):

    def __init__(self, df: Typing.PatchedDataFrame):
        self.df = df

    def feature_selection(self,
                          features_and_labels: FeaturesAndLabels,
                          top_features: int = 5,
                          correlation_threshold: float = 0.5,
                          minimum_features: int = 1,
                          lags: Iterable[int] = range(100),
                          show_plots: bool = True,
                          figsize: Tuple[int, int] = (12, 10),
                          **kwargs):
        # extract pandas objects
        kwargs = merge_kwargs(features_and_labels.kwargs, kwargs)
        (features, _), label, _, _, _ = extract_feature_labels_weights(self.df, features_and_labels, **kwargs)

        # try to estimate good features
        return feature_selection(features, label, top_features, correlation_threshold, minimum_features,
                                 lags, show_plots, figsize)

    def plot_features(self, data: Union[FeaturesAndLabels, MlModel]):
        fnl = data.features_and_labels if isinstance(data, MlModel) else data
        (features, _), labels, _, _, _ = self.df._.extract(fnl)
        return plot_features(features.join(labels), labels.columns[0] if has_indexed_columns(labels) else labels.name)

    def fit(self,
            model_provider: Callable[[], MlModel],
            training_data_splitter: Splitter = RandomSplits(),
            hyper_parameter_space: Dict = None,
            **kwargs
            ) -> Fit:
        return fit(self.df, model_provider, training_data_splitter, hyper_parameter_space, **kwargs)

    def backtest(self,
                 model: MlModel,
                 summary_provider: Callable[[Typing.PatchedDataFrame], Summary] = None,
                 **kwargs) -> Summary:
        return backtest(self.df, model, summary_provider, **kwargs)

    def predict(self,
                model: MlModel,
                tail: int = None,
                samples: int = 1,
                **kwargs) -> Typing.PatchedDataFrame:
        return predict(self.df, model, tail=tail, samples=samples, **kwargs)




from typing import Tuple, List, Generator

import numpy as np

from pandas_ml_utils.ml.data.splitting.sampeling import Sampler


def _get(arr, i):
    return arr[i] if arr is not None else None


class ExtractMultiMultiModelSampler(Sampler):

    def __init__(self, start_index, nr_models, sampler: Sampler):
        super().__init__(sampler.train, sampler.test, None)
        self.__cross_validation = sampler._cross_validation
        self.start_index = start_index
        self.nr_models = nr_models

    def sample(self) -> Generator[Tuple[List[np.ndarray], List[np.ndarray]], None, None]:
        cv = self.__cross_validation

        for s in super().sample():
            # features, labels, targets, weights, gross_loss
            (x, y, t, w, gl), (x_val, y_val, t_val, w_val, gl_val) = s
            nr_labels = y.shape[1] // self.nr_models
            # print(f'nr of features: {x.shape[1]}, nr of labels: {nr_labels}')

            def cut(arr):
                return arr[:, (nr_labels * self.start_index):(nr_labels * (self.start_index + 1))]

            _w, _w_val = None, None
            if w is not None:
                if w.ndim > 1:
                    if w.shape[1] > 1:
                        _w = cut(w)
                        _w_val = cut(w_val)
                    else:
                        _w = w[self.start_index]
                        _w_val = w_val[self.start_index]
                else:
                    _w = w
                    _w_val = w_val

            if cv is not None and isinstance(cv, Tuple) and callable(cv[1]):
                for fold_epoch in range(cv[0]):
                    # cross validation, make sure we re-shuffle every fold_epoch
                    y = cut(y)
                    for f, (train_idx, test_idx) in enumerate(cv[1](x, y)):
                        train = (x[train_idx], y[train_idx], _get(t, train_idx), _get(_w, train_idx), _get(gl, train_idx))
                        test = (x[test_idx], y[test_idx], _get(t, test_idx), _get(_w, test_idx), _get(gl, test_idx))
                        yield train, test
            else:
                # fit without cross validation
                yield (x, cut(y), t, _w, gl), (x_val, cut(y_val), t_val, _w_val, gl_val)



import matplotlib.pyplot as plt
from IPython.display import clear_output

from pandas_ml_utils.ml.summary.regression_summary import RegressionSummary


def plot_losses(parent):

    class PlotLosses(parent):
        def __init__(self):
            super().__init__()
            self.i = 0
            self.x = []
            self.losses = []
            self.val_losses = []
            self.fig = plt.figure()
            self.logs = []

        def on_epoch_end(self, epoch, logs={}):
            self.logs.append(logs)
            self.x.append(self.i)
            self.losses.append(logs.get('loss'))
            self.val_losses.append(logs.get('val_loss'))
            self.i += 1

            clear_output(wait=True)
            plt.plot(self.x, self.losses, label="loss")
            plt.plot(self.x, self.val_losses, label="val_loss")
            plt.legend()
            plt.show()

    return PlotLosses


def plot_true_pred_scatter(df, parent):

    class PlotTruePred(parent):
        def __init__(self):
            super().__init__()
            # FIXME
            #  1.) how to get train/test data
            #  2.) how to get a frame with label/prediction columns
            self.sum = RegressionSummary(df)

        def on_epoch_end(self, epoch, logs={}):
            clear_output(wait=True)
            self.sum.plot_true_pred_scatter()

    raise NotImplementedError() # TODO implement me ...
    return PlotTruePred




from typing import Tuple

from matplotlib.axis import Axis
from matplotlib.figure import Figure

from pandas_ml_common import Typing
from pandas_ml_common.utils.serialization_utils import plot_to_html_img
from pandas_ml_utils import html
from pandas_ml_utils.constants import *
from pandas_ml_utils.ml.summary import Summary
from pandas_ml_utils.ml.model.base_model import Model


class RegressionSummary(Summary):

    def __init__(self, df: Typing.PatchedDataFrame, model: Model, true_columns=None, pred_columns=None, **kwargs):
        super().__init__(df, model, **kwargs)
        self.true_columns = LABEL_COLUMN_NAME if true_columns is None else true_columns
        self.pred_columns = PREDICTION_COLUMN_NAME if pred_columns is None else pred_columns

    def plot_true_pred_scatter(self, figsize=(6, 6), alpha=0.1) -> Tuple[Figure, Axis]:
        import matplotlib.pyplot as plt

        df = self.df
        x = df[self.pred_columns]._.values
        y = df[self.true_columns]._.values
        mn = min(x.min(), y.min())
        mx = max(x.max(), y.max())

        fig, axis = plt.subplots(figsize=figsize)
        plt.axes(aspect='equal')
        plt.xlabel("Predicted")
        plt.ylabel("True")

        plt.scatter(x, y, alpha=alpha)
        plt.xlim([mn, mx])
        plt.ylim([mn, mx])
        plt.plot([mn, mx], [mn, mx])

        return fig, axis

    def __str__(self):
        return f"... to be implemented ..."

    def _repr_html_(self):
        from mako.template import Template
        from mako.lookup import TemplateLookup

        template = Template(filename=html.SELF_TEMPLATE(__file__), lookup=TemplateLookup(directories=['/']))
        return template.render(true_pred_scatter_plot=plot_to_html_img(self.plot_true_pred_scatter))


import os
import tempfile
import uuid
from unittest import TestCase

import numpy as np
import keras
from keras.layers import Dense
from keras.models import Sequential
from keras.optimizers import SGD
from tensorflow import set_random_seed

from pandas_ml_utils import KerasModel, FeaturesAndLabels, Model
from pandas_ml_utils_test.ml.data.model.test_abstract_model import TestAbstractModel
from pandas_ml_common import pd

os.environ["CUDA_VISIBLE_DEVICES"] = ""
set_random_seed(42)
np.random.seed(42)


def custom_loss_function(y_true, y_pred):
    return keras.losses.mean_squared_error(y_true, y_pred) * keras.backend.constant(2)


class TestKerasModel(TestAbstractModel, TestCase):

    def provide_classification_model(self, features_and_labels):
        def model_provider():
            model = Sequential([
                Dense(3, input_shape=(2,), activation='tanh'),
                Dense(1, activation='sigmoid')
            ])

            model.compile(optimizer=SGD(lr=0.1, momentum=0.9), loss='mean_squared_error')
            return model

        model = KerasModel(
            model_provider,
            features_and_labels,
        )

        return model

    def provide_regression_model(self, features_and_labels):
        def model_provider():
            model = Sequential([
                Dense(units=1, input_shape=(1, ))
            ])

            model.compile(optimizer='sgd', loss='mean_squared_error')
            return model

        model = KerasModel(
            model_provider,
            features_and_labels,
        )

        return model

    def test_custom_objects(self):
        df = pd.DataFrame({
            "a": [-1.0, 0.0, 1.0, 2.0, 3.0, 4.0],
            "b": [-2.0, 1.0, 4.0, 7.0, 10.0, 13.0]
        })

        def model_provider():
            model = Sequential([
                Dense(units=1, input_shape=(1,))
            ])

            model.compile(optimizer='sgd', loss=custom_loss_function)
            return model, custom_loss_function

        model = KerasModel(
            model_provider,
            FeaturesAndLabels(["a"], ["b"]),
        )

        fit = df.model.fit(model)
        temp = os.path.join(tempfile.gettempdir(), str(uuid.uuid4()))
        try:
            fit.model.save(temp)
            copy = Model.load(temp)
            pd.testing.assert_frame_equal(df.model.predict(fit.model), df.model.predict(copy), check_less_precise=True)
        finally:
            os.remove(temp)






from unittest import TestCase

import torch.nn as nn
from torch.optim import Adam

from pandas_ml_common.decorator import MultiFrameDecorator
from pandas_ml_utils import pd
from pandas_ml_utils.ml.model.pytoch_model import PytorchModel
from pandas_ml_utils import FeaturesAndLabels


class TestMultiFeatureSet(TestCase):

    def test_pytorch_mfs(self):
        df = pd.DataFrame({
            "a": [1, 0, 1, 0, 1, 0, 1, 0,],
            "b": [[0, 0], [0, 0], [1, 1], [1,1], [0,0], [0,0], [1,1], [1,1],],
            "c": [1, 0, 0, 1, 1, 0, 0, 1,]
        })

        def module_provider():
            class ClassificationModule(nn.Module):

                def __init__(self):
                    super().__init__()
                    self.classifier = nn.Sequential(
                        nn.Linear(2, 5),
                        nn.ReLU(),
                        nn.Linear(5, 1),
                        nn.Sigmoid()
                    )

                def forward(self, x):
                    x = self.classifier(x)
                    return x

            return ClassificationModule()

        model = PytorchModel(
            FeaturesAndLabels(
                features=(["a"], ["b"]),
                labels=["c"]
            ),
            module_provider,
            nn.MSELoss,
            lambda params: Adam(params, lr=0.03)
        )

        (f, _), l, t, s, g = df._.extract(model.features_and_labels)
        self.assertIsInstance(f, MultiFrameDecorator)
        print(f)

#        fit = df.model.fit(model)
        # FIXME print(fit)



from __future__ import annotations

import contextlib
import logging
import os
import tempfile
import uuid
from copy import deepcopy
from typing import List, Callable, TYPE_CHECKING, Tuple

import numpy as np

from pandas_ml_common import Typing
from pandas_ml_common.utils import merge_kwargs, suitable_kwargs
from pandas_ml_utils.ml.data.extraction import FeaturesAndLabels
from pandas_ml_utils.ml.summary import Summary
from .base_model import Model

_log = logging.getLogger(__name__)


class KerasModel(Model):
    # eventually we need to save and load the weights of the keras model individually by using `__getstate__`
    #  `__setstate__` like described here: http://zachmoshe.com/2017/04/03/pickling-keras-models.html
    if TYPE_CHECKING:
        from keras.models import Model as KModel

    def __init__(self,
                 keras_compiled_model_provider: Callable[[], KModel],
                 features_and_labels: FeaturesAndLabels,
                 output_shape: Tuple[int,...] = (-1,),
                 summary_provider: Callable[[Typing.PatchedDataFrame], Summary] = Summary,
                 callbacks: List[Callable] = [],
                 **kwargs):
        """
        Keras compatible implementation of :class:`.Model`.
        :param keras_compiled_model_provider: a callable which provides an eventually compiled ready to fit keras model.
               if the model is not compiled you ned to pass "optimizer" argument as kwargs.
               NOTE: the tensorflow backend is currently limited to 1.*
        :param features_and_labels: see :class:`.Model`
        :param summary_provider: :class:`.Model`.
        :param epochs: number of epochs passed to the keras fit function
        :param callbacks: a list of callable's providing a keras compatible callback. It is neccessary to pass the
                          callable i.e. BaseLogger or lambda: BaseLogger(...) vs. BaseLogger(...)
        :param kwargs: a list of arguments passed to the keras_compiled_model_provider and keras fit function as they
                       match the signature
        """
        super().__init__(features_and_labels, summary_provider, **kwargs)
        self.keras_model_provider = keras_compiled_model_provider
        self.output_shape = output_shape
        self.custom_objects = {}

        import keras
        if keras.backend.backend() == 'tensorflow':
            import tensorflow as tf
            self.is_tensorflow = True
            self.graph = tf.Graph()
            with self.graph.as_default():
                self.session = tf.Session(graph=self.graph)
        else:
            self.is_tensorflow = False

        # create keras model
        provider_args = suitable_kwargs(keras_compiled_model_provider, **kwargs)
        keras_model = self._exec_within_session(keras_compiled_model_provider, **provider_args)

        if isinstance(keras_model, Tuple):
            # store custom objects
            for i in range(1, len(keras_model)):
                if hasattr(keras_model[i], "__name__"):
                    self.custom_objects[keras_model[i].__name__] = keras_model[i]
                else:
                    raise ValueError("keras custom object must have a __name__")

            keras_model = keras_model[0]
            _log.warning(f"keras model using custom objects {self.custom_objects}")

        # eventually compile keras model
        if not keras_model.optimizer:
            compile_args = suitable_kwargs(keras_model.compile, **kwargs)
            self._exec_within_session(keras_model.compile, **compile_args)

        # set all members
        self.keras_model = keras_model
        self.callbacks = callbacks
        self.history = None

    def fit_fold(self,
                 fold_nr: int,
                 x: np.ndarray, y: np.ndarray,
                 x_val: np.ndarray, y_val: np.ndarray,
                 sample_weight_train: np.ndarray, sample_weight_test: np.ndarray,
                 **kwargs) -> Tuple[np.ndarray, np.ndarray]:
        fitter_args = suitable_kwargs(self.keras_model.fit, {"callbacks": []}, self.kwargs, kwargs)
        fitter_args["callbacks"] = [*fitter_args["callbacks"], *[cb() for cb in self.callbacks]]

        if sample_weight_train is not None:
            print(f"using sample weights {sample_weight_train.shape} -> (?, 1)")
            sample_weight_train = sample_weight_train.ravel()

        if len(fitter_args) > 0:
            print(f'pass args to fit: {fitter_args}')

        fit_history = self._exec_within_session(self.keras_model.fit,
                                                x,
                                                y.reshape(y.shape[:1] + self.output_shape),
                                                sample_weight=sample_weight_train,
                                                validation_data=(
                                                    x_val,
                                                    y_val.reshape(y_val.shape[:1] + self.output_shape)
                                                ),
                                                **fitter_args)
        if self.history is None:
            self.history = fit_history.history
        else:
            for metric, _ in self.history.items():
                self.history[metric] = self.history[metric] + fit_history.history[metric]

        return np.array(fit_history.history['loss']), np.array(fit_history.history['val_loss'])

    def predict_sample(self, x: np.ndarray, **kwargs) -> np.ndarray:
        return self._exec_within_session(self.keras_model.predict, x)

    def get_weights(self):
        return self._exec_within_session(self.keras_model.get_weights)

    def set_weights(self, weights):
        self._exec_within_session(self.keras_model.set_weights, weights)

    def get_tensor_by_name(self, i):
        if self.is_tensorflow:
            with self.graph.as_default():
                with self.session.as_default():
                    #return self.graph.get_tensor_by_name(name)
                    return self.keras_model.get_layer(None, i)

    def _exec_within_session(self, func, *args, **kwargs):
        if self.is_tensorflow:
            with self.graph.as_default():
                with self.session.as_default():
                    return func(*args, **kwargs)
        else:
            return func(*args, **kwargs)

    def __getstate__(self):
        # Copy the object's state from self.__dict__ which contains all our instance attributes.
        # Always use the dict.copy() method to avoid modifying the original state.
        state = self.__dict__.copy()

        # remove un-pickleable fields
        if self.is_tensorflow:
            del state['graph']
            del state['session']

        # special treatment for the keras model
        tmp_keras_file = os.path.join(tempfile.gettempdir(), str(uuid.uuid4()))
        self._exec_within_session(lambda: self.keras_model.save(tmp_keras_file, True, True))
        with open(tmp_keras_file, mode='rb') as file:
            state['keras_model'] = file.read()

        # clean up temp file
        with contextlib.suppress(OSError):
            os.remove(tmp_keras_file)

        # return state
        return state

    def __setstate__(self, state):
        from keras.models import load_model

        # restore keras file
        tmp_keras_file = os.path.join(tempfile.gettempdir(), str(uuid.uuid4()))
        with open(tmp_keras_file, mode='wb') as file:
            file.write(state['keras_model'])
            del state['keras_model']

        # Restore instance attributes
        self.__dict__.update(state)

        # restore keras model and tensorflow session if needed
        if self.is_tensorflow:
            from keras import backend as K
            import tensorflow as tf
            self.graph = tf.Graph()
            with self.graph.as_default():
                self.session = tf.Session(graph=self.graph)
                K.set_session(self.session)
                self.keras_model = load_model(tmp_keras_file, custom_objects=self.custom_objects)
        else:
            self.keras_model = load_model(tmp_keras_file, custom_objects=self.custom_objects)

        # clean up temp file
        with contextlib.suppress(OSError):
            os.remove(tmp_keras_file)

    def __del__(self):
        import keras.backend as K

        if self.is_tensorflow:
            try:
                K.clear_session()
                self.session.close()
            except AttributeError:
                _log.warning(f"failed to clear keras session!")

    def __call__(self, *args, **kwargs):
        new_model = KerasModel(self.keras_model_provider,
                               self.features_and_labels,
                               self.output_shape,
                               self.summary_provider,
                               deepcopy(self.callbacks),
                               **merge_kwargs(deepcopy(self.kwargs), kwargs))

        # copy weights before return
        new_model.set_weights(self.get_weights())

        return new_model




from unittest import TestCase

from pandas_ml_utils import pd, np
from pandas_ml_utils.ml.data.splitting import RandomSplits


class TestTrainTestData(TestCase):

    def test_no_training_data(self):
        """given"""
        df = pd.DataFrame({"featureA": [1,2,3,4,5],
                           "labelA": [1,2,3,4,5]})

        """when"""
        train_ix, test_ix = RandomSplits(0).train_test_split(df.index)

        """then"""
        np.testing.assert_array_almost_equal(train_ix.values, df.index.values)
        self.assertEqual(0, len(test_ix))

    def test_make_training_data(self):
        """given"""
        df = pd.DataFrame({"featureA": [1, 2, 3, 4, 5],
                           "labelA": [1, 2, 3, 4, 5]})

        """when"""
        train_ix, test_ix = RandomSplits(test_size=0.5).train_test_split(df.index)

        """then"""
        self.assertEqual(2, len(train_ix))
        self.assertEqual(3, len(test_ix))

    def test_youngest_portion(self):
        """given"""
        df = pd.DataFrame({"featureA": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
                           "labelA": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]})

        """when"""
        train_ix, test_ix = RandomSplits(test_size=0.6, youngest_size=0.25).train_test_split(df.index)

        "then"
        self.assertEqual(6, len(test_ix))
        np.testing.assert_array_equal(test_ix[-2:], np.array([8, 9]))



"""Augment pandas DataFrame with methods for machine learning"""
__version__ = '0.1.14.1'

from pandas_ml_utils.ml.summary import *
from pandas_ml_common import *
from pandas_ml_utils.df.model import Model as DfModelExtension
from pandas_ml_utils.ml.data import *
from pandas_ml_utils.ml.data.extraction import FeaturesAndLabels
from pandas_ml_utils.ml.model import *

_log = logging.getLogger(__name__)
_log.debug(f"numpy version {np.__version__}")
_log.debug(f"pandas version {pd.__version__}")


setattr(PandasObject, "model", property(lambda self: DfModelExtension(self)))



from typing import Tuple, Callable

import pandas as pd
import numpy as np

from pandas_ml_common.utils.random import temp_seed
from pandas_ml_utils.ml.data.splitting.splitter import Splitter


class RandomSequences(Splitter):

    def __init__(self, test_size=0.4, session_size=0.7, max_folds=100, seed=None):
        super().__init__()
        self.test_size = test_size
        self.session_size = session_size
        self.max_folds = max_folds
        self.seed = seed
        self.min_fold_validation_size=2
        self.min_training_samples=2

    def train_test_split(self, index: pd.Index) -> Tuple[pd.Index, pd.Index]:
        # we just split the sequence int past and recent data
        end_idx = int(len(index) * (1 - self.test_size))
        return index[0:end_idx], index[end_idx:]

    @property
    def cross_validation(self) -> Tuple[int, Callable[[pd.Index, pd.Index], Tuple[np.ndarray, np.ndarray]]]:
        # this is the magic part of this splitter, because we randomly start from the taining set only moving forward
        # if max_folds is None then we keep indefinitely sampling data

        def sampler(features_index, labels_index) -> Tuple[np.ndarray, np.ndarray]:
            # convert pandas index to array index
            index = np.arange(len(features_index))

            # calculate the latest possible index such that we can sample a whole session
            max_idx = int(len(index) * (1 - self.session_size)) - self.min_fold_validation_size - self.min_training_samples
            assert max_idx > 0, f"not enough data! {features_index.shape}"

            # sample the data (make sure we have at least one sample in the training data)
            for i in (range(self.max_folds) if self.max_folds is not None else Splitter.infinity_sample_range()):
                if self.seed is not None:
                    with temp_seed(self.seed):
                        idx = np.random.choice(max_idx)
                else:
                    idx = np.random.choice(max_idx)

                # finally yield the resulting indices
                yield index[:idx+self.min_training_samples], index[idx+self.min_training_samples:]

        return 1, sampler




import unittest

from pandas_ml_utils import pd, FeaturesAndLabels


class TestFeatureSelection(unittest.TestCase):

    def test_feature_selection(self):
        df = pd.DataFrame({"featureA": [1, 2, 3, 4, 5],
                           "featureB": [5, 4, 3, 2, 1],
                           "featureC": [1, 2, 1, 2, 1],
                           "labelA": [1, 2, 3, 4, 5],
                           "labelB": [5, 4, 3, 2, 1]})


        analysis = df.model.feature_selection(FeaturesAndLabels(["featureA", "featureB", "featureC"], ["labelA"]),
                                              lags=[2], show_plots=False)


        print(analysis)
        # top features are A, B, C
        self.assertListEqual(["featureA", "featureB", "featureC"], analysis[0])
        self.assertListEqual([0, 1], analysis[1])


if __name__ == '__main__':
    unittest.main()



import logging
from typing import Tuple, Callable

import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split as sk_train_test_split

from pandas_ml_utils.ml.data.splitting.splitter import Splitter

_log = logging.getLogger(__name__)


class RandomSplits(Splitter):

    def __init__(self,
                 test_size: float = 0.4,
                 youngest_size: float = None,
                 cross_validation: Tuple[int, Callable[[pd.Index, pd.Index], Tuple[np.ndarray, np.ndarray]]] = None,
                 test_validate_split_seed=42):
        """
        :param test_size: the fraction [0, 1] of random samples which are used for a test set
        :param youngest_size: the fraction [0, 1] of the test samples which are not random but are the youngest
        :param cross_validation: tuple of number of epochs for each fold provider and a cross validation provider
        :param test_validate_split_seed: seed if train, test splitting needs to be reproduceable. A magic seed 'youngest' is
                                         available, which just uses the youngest data as test data
        """

        super().__init__()
        self.test_size = test_size
        self.youngest_size = test_size * youngest_size if youngest_size is not None else None
        self._cross_validation = cross_validation
        self.seed = test_validate_split_seed

    def train_test_split(self, index: pd.Index) -> Tuple[pd.Index, pd.Index]:

        # convert data frame index to numpy array
        index = index.values

        if self.test_size <= 0:
            train, test = index, index[:0]
        elif self.seed == 'youngest':
            i = int(len(index) - len(index) * self.test_size)
            train, test = index[:i], index[i:]
        else:
            random_sample_test_size = self.test_size if self.youngest_size is None else self.test_size * (1 - self.youngest_size)
            random_sample_train_index_size = int(len(index) - len(index) * (self.test_size - random_sample_test_size))

            if random_sample_train_index_size < len(index):
                _log.warning(f"keeping youngest {len(index) - random_sample_train_index_size} elements in test set")

                # cut the youngest data and use residual to randomize train/test data
                index_train, index_test = \
                    sk_train_test_split(index[:random_sample_train_index_size],
                                     test_size=random_sample_test_size, random_state=self.seed)

                # then concatenate (add back) the youngest data to the random test data
                index_test = np.hstack([index_test, index[random_sample_train_index_size:]])  # index is 1D

                train, test = index_train, index_test
            else:
                train, test = sk_train_test_split(index, test_size=random_sample_test_size, random_state=self.seed)

        return pd.Index(train), pd.Index(test)

    @property
    def cross_validation(self) -> Tuple[int, Callable[[pd.Index, pd.Index], Tuple[np.ndarray, np.ndarray]]]:
        return self._cross_validation




from typing import Tuple

import numpy as np
from matplotlib.axis import Axis
from matplotlib.figure import Figure
from mlxtend.evaluate import confusion_matrix
from sklearn.metrics import roc_curve, auc

from pandas_ml_common import Typing
from pandas_ml_common.utils.serialization_utils import plot_to_html_img
from pandas_ml_utils import html
from pandas_ml_utils.constants import *
from pandas_ml_utils.ml.model.base_model import Model
from pandas_ml_utils.ml.summary import Summary


class ClassificationSummary(Summary):

    def __init__(self, df: Typing.PatchedDataFrame, model: Model, true_columns=None, pred_columns=None, **kwargs):
        super().__init__(df, model, **kwargs)
        self.true_columns = LABEL_COLUMN_NAME if true_columns is None else true_columns
        self.pred_columns = PREDICTION_COLUMN_NAME if pred_columns is None else pred_columns

    def plot_ROC(self, figsize=(6, 6)) -> Tuple[Figure, Axis]:
        import matplotlib.pyplot as plt

        # get true and prediction data
        truth = self.df[self.true_columns]._.values.reshape((len(self.df), -1))
        prediction = self.df[self.pred_columns]._.values.reshape(truth.shape)

        # fix binary classification case
        if truth.shape[-1] == 1:
            truth = np.hstack([truth, 1 - truth])

        if prediction.shape[-1] == 1:
            prediction = np.hstack([prediction, 1 - prediction])

        # calculate legends
        legend = [(col[1] if isinstance(col, tuple) else col) for col in self.df[self.true_columns].columns.tolist()]
        if truth.shape[1] > len(legend):
            legend = list(range(truth.shape[1]))

        # Compute ROC curve and ROC area for each class
        fpr = dict()
        tpr = dict()
        roc_auc = dict()

        if truth.ndim > 1:
            for i in range(truth.shape[1]):
                fpr[i], tpr[i], _ = roc_curve(truth[:, i], prediction[:, i])
                roc_auc[i] = auc(fpr[i], tpr[i])
        else:
            fpr[0], tpr[0], _ = roc_curve(truth, prediction)
            roc_auc[0] = auc(fpr[0], tpr[0])

        # plot ROC curves
        fig, axis = plt.subplots(figsize=figsize)

        for i in fpr.keys():
            plt.plot(fpr[i], tpr[i], label=f"{legend[i]} auc:{roc_auc[i] * 100:.2f}")

        plt.plot([0, 1], [0, 1], color='navy', linestyle='--')
        plt.xlim([0.0, 1.0])
        plt.ylim([0.0, 1.05])
        plt.xlabel('False Positive Rate')
        plt.ylabel('True Positive Rate')
        plt.title('Receiver Operating Characteristic')
        plt.legend(loc="lower right")
        return fig, axis

    def plot_confusion_matrix(self, figsize=(6, 6)) -> Tuple[Figure, Axis]:
        from mlxtend.plotting import plot_confusion_matrix

        truth, prediction = self._fix_label_prediction_representation()
        distinct_values = {*truth.reshape((-1,))}

        cm = confusion_matrix(truth, prediction, binary=len(distinct_values) <= 2)
        fig, ax = plot_confusion_matrix(cm, figsize=figsize)
        # ax.set_title('Confusion Matrix', fontsize=1)

        return fig, ax

    def _fix_label_prediction_representation(self):
        true_values = self.df[self.true_columns]._.values

        if true_values.ndim > 1:
            true_values = true_values.reshape((len(true_values), -1))

        pred_values = self.df[self.pred_columns]._.values.reshape(true_values.shape)

        if true_values.ndim > 1 and true_values.shape[1] > 2:
            # get class of multi class probabilities
            true_values = np.apply_along_axis(np.argmax, 1, true_values)
            pred_values = np.apply_along_axis(np.argmax, 1, pred_values)
        elif pred_values.shape[1] == 1:
            # fix for binary classification case
            pred_values = pred_values > 0.5

        return true_values.reshape((-1, 1)), pred_values.reshape((-1, 1))

    def __str__(self):
        truth, prediction = self._fix_label_prediction_representation()
        distinct_values = {*truth.reshape((-1,))}

        cmx = confusion_matrix(truth, prediction, binary=len(distinct_values) <= 2)
        return f"{cmx}"

    def _repr_html_(self):
        from mako.template import Template
        from mako.lookup import TemplateLookup

        template = Template(filename=html.SELF_TEMPLATE(__file__), lookup=TemplateLookup(directories=['/']))
        return template.render(fit=self,
                               roc_plot=plot_to_html_img(self.plot_ROC),
                               cmx_plot=plot_to_html_img(self.plot_confusion_matrix))



from typing import Tuple, Callable

import numpy as np
import pandas as pd

from pandas_ml_utils.ml.data.splitting.splitter import Splitter


class DummySplitter(Splitter):

    def __init__(self, samples_size=1):
        super().__init__()
        self.sample_size = samples_size

    def train_test_split(self, index: pd.Index) -> Tuple[pd.Index, pd.Index]:
        return index, index[[]]

    @property
    def cross_validation(self) -> Tuple[int, Callable[[pd.Index, pd.Index], Tuple[np.ndarray, np.ndarray]]]:
        def just_repeat(i1, i2):
            idx = np.arange(len(i1))
            for _ in range(self.sample_size):
                yield idx, idx

        return 1, just_repeat



import logging
from typing import Callable, Tuple, Generator, List, Dict, Any

import gym
import numpy as np
import pandas as pd

from pandas_ml_common import Typing
from pandas_ml_common.utils import call_callable_dynamic_args
from pandas_ml_utils.ml.data.extraction import FeaturesAndLabels
from pandas_ml_utils.ml.summary import Summary
from .base_model import Model
from ..data.splitting.sampeling import Sampler

_log = logging.getLogger(__name__)


class ReinforcementModel(Model):

    class RLModel(object):
        def learn(self, total_timesteps, callback=None,
                  log_interval=100, tb_log_name="run", reset_num_timesteps=True):
            pass

        def predict(self, observation, state=None, mask=None, deterministic=False):
            pass

        def get_env(self):
            pass

        def set_env(self, env):
            pass

    class DataFrameGym(gym.Env):

        def __init__(self,
                     action_space: gym.spaces.Space,
                     observation_space: gym.spaces.Space):
            super().__init__()
            self.action_space = action_space
            self.observation_space = observation_space

            # initialize palace holder variables
            # TODO exclude these variables from serialization
            self.reward_history_collector = None
            self.data_generator: Generator[Tuple[List[np.ndarray], List[np.ndarray]], None, None] = None
            self.last_reward: float = None
            self.current_index: int = 0
            self.last_index: int = -1
            self.train: Tuple[List[np.ndarray]] = None
            # self.test: Tuple[List[np.ndarray]] = None

        def reset(self) -> np.ndarray:
            # get a new data set from the generator
            # NOTE the total numbers which can be drawn from the generator must be greater or equal to the total number
            # of iteration of the learning agent
            self.train, _ = next(self.data_generator)

            # reset indices and max timesteps and add a new reward history
            self.current_index = 0
            self.last_index = len(self.train[0])
            self.reward_history_collector(None)

            # return the very first observation
            i = self.current_index
            return self.next_observation(i, *[t[i] if t is not None else None for t in self.train])

        def step(self, action) -> Tuple[np.ndarray, float, bool, Dict]:
            # initialize reward for this action
            done = False
            reward = 0

            # try to execute the action, if it fails with an exception we are done with this session
            try:
                i = self.current_index
                state = [t[i] if t is not None else None for t in self.train]
                interpreted_action = self.interpret_action(action, i, *state)
                reward = self.take_action(interpreted_action, i, *state)
                self.reward_history_collector(reward)
                self.last_reward = reward
            except StopIteration:
                done = True

            # now update the observation of the environment
            self.current_index += 1
            i_new = self.current_index
            done = done or i_new >= self.last_index
            observation = self.next_observation(i_new, *[t[i_new] if t is not None else None for t in self.train]) if not done else None

            # return the new observation the reward of this action and whether the session is over
            return observation, reward, done, {"interpreted_action": interpreted_action}

        def interpret_action(self,
                        action,
                        idx: int,
                        features: np.ndarray,
                        labels: np.ndarray,
                        targets: np.ndarray,
                        weights: np.ndarray = None,
                        gross_loss: np.ndarray = None) -> float:
            return action

        def take_action(self,
                        action,
                        idx: int,
                        features: np.ndarray,
                        labels: np.ndarray,
                        targets: np.ndarray,
                        weights: np.ndarray = None,
                        gross_loss: np.ndarray = None) -> float:
            pass

        def next_observation(self,
                             idx: int,
                             features: np.ndarray,
                             labels: np.ndarray,
                             targets: np.ndarray,
                             weights: np.ndarray = None,
                             gross_loss: np.ndarray = None) -> np.ndarray:
            pass

        def set_data_generator(self, sampler: Sampler, reward_history_collector, **kwargs):
            self.data_generator = sampler.sample()
            self.reward_history_collector = reward_history_collector
            return self.reset()

    def __init__(self,
                 reinforcement_model_provider: Callable[[Any], RLModel],
                 features_and_labels: FeaturesAndLabels,
                 summary_provider: Callable[[Typing.PatchedDataFrame], Summary] = Summary,
                 **kwargs):
        super().__init__(features_and_labels, summary_provider, **kwargs)
        self.reinforcement_model_provider = reinforcement_model_provider
        self.rl_model = call_callable_dynamic_args(reinforcement_model_provider, **self.kwargs, **kwargs)
        self.reward_history = None
        self.env = self.rl_model.get_env()

    def fit(self, sampler: Sampler, **kwargs) -> float:
        # provide the data generator for every environment
        envs = self.rl_model.get_env().envs
        self.reward_history = []

        for env in envs:
            rh = []
            self.reward_history.append(rh)
            env.set_data_generator(sampler,
                                   lambda reward: rh[-1].append(reward) if reward is not None else rh.append([]),
                                   **self.kwargs)

        call_callable_dynamic_args(self.rl_model.learn, **self.kwargs, **kwargs)

        # collect statistics of all environments
        latest_rewards = np.array([rh[-1] for erh in self.reward_history for rh in erh if len(rh) > 1])
        return latest_rewards.mean()

    def predict(self, sampler: Sampler, **kwargs) -> np.ndarray:
        # set a fresh env
        self.rl_model.set_env(self.env)
        samples, _ = sampler.nr_of_source_events
        envs = self.rl_model.get_env().envs[:1]
        rh = []

        obs = [env.set_data_generator(sampler,
                                      lambda reward: rh[-1].append(reward) if reward is not None else rh.append([]),
                                      **self.kwargs) for env in envs]

        prediction = []
        done = False

        for i in range(samples):
            if not done:
                action, state = self.rl_model.predict(obs)
                obs, reward, done, info = zip(*[env.step(action) for env in envs])
                prediction.append(info[0]["interpreted_action"])

                for env, dne in zip(envs, done):
                    call_callable_dynamic_args(env.render, kwargs)

                done = any(done)
            else:
                prediction.append(np.nan)

        return np.array(prediction)

    def plot_loss(self, accumulate_reward=False):
        import matplotlib.pyplot as plt
        plots = 0
        mins = []

        for env_hist in self.reward_history:
            for i, session in enumerate(env_hist):
                if len(session) > 1:
                    s = pd.Series(session)
                    if accumulate_reward:
                        s = s.cumsum()

                    mins.append(s.min())
                    plt.plot(s, label=f'reward {i}')
                    plots += 1

        # only plot legend if it fits on the plot
        if plots <= 10:
            plt.legend(loc='best')

        print(f'worst reward: {np.array(mins).min()}')

    def __getstate__(self):
        # FIXME need to be implemented
        # rl_model.get_env() -> save the first env!
        pass

    def __setstate__(self, state):
        # FIXME need to be implemented
        # rl_model.set_env(saved <- envs)
        pass

    def __call__(self, *args, **kwargs):
        # create a new version of this model
        return ReinforcementModel(
            self.reinforcement_model_provider,
            self.features_and_labels,
            self.summary_provider,
            **self.kwargs, **kwargs
        )



from unittest import TestCase

from sklearn.neural_network import MLPClassifier, MLPRegressor

from pandas_ml_common import np, pd
from pandas_ml_utils import SkModel, FeaturesAndLabels, MultiModel, MultiModelSummary, ClassificationSummary
from pandas_ml_utils.constants import PREDICTION_COLUMN_NAME
from pandas_ml_utils.ml.data.splitting import NaiveSplitter


class TestSkMultiModel(TestCase):

    def test_multi_model(self):
        """given some toy classification data"""
        df = pd.DataFrame({
            "a": [1, 0, 1, 0, 1, 0, 1, 0,],
            "b": [0, 0, 1, 1, 0, 0, 1, 1,],
            "c": [1, 0, 0, 1, 1, 0, 0, 1,],
        })

        model = MultiModel(
            SkModel(
                MLPClassifier(activation='logistic', max_iter=1000, hidden_layer_sizes=(3,), alpha=0.001,
                              solver='lbfgs', random_state=42),
                FeaturesAndLabels(
                    features=["a", "b"],
                    labels=[lambda df, i: df["c"].rename(f"c_{i}")],
                    label_type=int),
                summary_provider=ClassificationSummary
            ),
            2,
            model_index_variable="i",
            summary_provider=MultiModelSummary
        )

        fit = df.model.fit(model, NaiveSplitter(0.49), epochs=1500, verbose=True)
        print(fit.training_summary._repr_html_()[:100])

        pdf = df.model.predict(fit.model, tail=2)
        print(pdf)

    def test_multi_model_multi_label(self):
        """given some toy classification data"""
        df = pd.DataFrame({
            "a": [1, 0, 1, 0, 1, 0, 1, 0,],
            "b": [0, 0, 1, 1, 0, 0, 1, 1,],
            "c": [1, 0, 0, 1, 1, 0, 0, 1,],
            "d": [1, 0, 0, 1, 1, 0, 0, 1,],
        })

        model = MultiModel(
            SkModel(
                MLPRegressor(activation='logistic', max_iter=1000, hidden_layer_sizes=(3,), alpha=0.001,
                             solver='lbfgs', random_state=42),
                FeaturesAndLabels(
                    features=["a", "b"],
                    labels=[lambda df, i: df["c"].rename(f"c_{i}"), lambda df, i: df["d"].rename(f"d_{i}")],
                    label_type=int),
                summary_provider=ClassificationSummary
            ),
            2,
            model_index_variable="i",
            summary_provider=MultiModelSummary
        )

        fit = df.model.fit(model, NaiveSplitter(0.49), epochs=1500, verbose=True)
        print(fit.training_summary._repr_html_()[:100])

        self.assertEqual(4, len(fit.training_summary.df[PREDICTION_COLUMN_NAME, "c_0"]))
        self.assertEqual(4, len(fit.training_summary.df[PREDICTION_COLUMN_NAME, "c_1"]))
        np.testing.assert_array_almost_equal(
            fit.training_summary.df[PREDICTION_COLUMN_NAME, "c_0"],
            fit.training_summary.df[PREDICTION_COLUMN_NAME, "c_1"]
        )

        pdf = df.model.predict(fit.model, tail=2)
        print(pdf)

    def test_multi_model_kwargs(self):
        """given some toy classification data"""
        df = pd.DataFrame({
            "a": [1, 0, 1, 0, 1, 0, 1, 0,],
            "b": [0, 0, 1, 1, 0, 0, 1, 1,],
            "c": [1, 0, 0, 1, 1, 0, 0, 1,],
        })

        model = MultiModel(
            SkModel(
                MLPClassifier(activation='logistic', max_iter=1000, hidden_layer_sizes=(3,), alpha=0.001,
                              solver='lbfgs', random_state=42),
                FeaturesAndLabels(
                    features=["a", "b"],
                    labels=[lambda df, index: df["c"].rename(index)],
                    label_type=int,
                    index=["z"]
                ),
            ),
            ["c1", "c2"],
            model_index_variable="index"
        )

        fit = df.model.fit(model, NaiveSplitter(0.49), epochs=1500, verbose=True)
        print(fit.training_summary.df)

        self.assertEqual(4, len(fit.training_summary.df[PREDICTION_COLUMN_NAME, "c1"]))
        self.assertEqual(4, len(fit.training_summary.df[PREDICTION_COLUMN_NAME, "c2"]))
        np.testing.assert_array_almost_equal(
            fit.training_summary.df[PREDICTION_COLUMN_NAME, "c1"],
            fit.training_summary.df[PREDICTION_COLUMN_NAME, "c2"]
        )




from typing import Union, Iterable

import numpy as np
import pandas as pd
from sklearn.preprocessing import OneHotEncoder, LabelEncoder
from sklearn.neural_network import MLPClassifier, MLPRegressor
from sklearn.linear_model import LinearRegression, LogisticRegression
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor
from sklearn.metrics import r2_score, f1_score
from sklearn.model_selection import KFold
from sklearn.model_selection._split import _BaseKFold

from pandas_ml_common import Typing as t
from pandas_ml_utils.ml.data.selection.feature_selection_report import FeatureSelectionReport, FeatureScoreReport


LINEAR = {
    "regressor": LinearRegression(),
    "classifier": LogisticRegression(multi_class='multinomial')
}

TREE = {
    "regressor": DecisionTreeRegressor(),
    "classifier": DecisionTreeClassifier()
}

MLP = {
    "regressor": MLPRegressor((30, 10), activation='tanh', early_stopping=True),
    "classifier": MLPClassifier((30, 10), activation='tanh', early_stopping=True)
}



def autoselect_features(df: t.PatchedDataFrame) -> FeatureSelectionReport:
    # TODO implement me
    return FeatureSelectionReport()


def score_feature(
        df,
        label,
        features,
        lags: Union[int, Iterable[int]] = [0],
        cv: _BaseKFold = KFold(6),
        regressor=DecisionTreeRegressor(),
        classifier=DecisionTreeClassifier()
    ) -> FeatureScoreReport:
    import matplotlib.pyplot as plt

    # get X and y data
    _X = df._[features]
    _y = df._[label]

    # convert categorical data
    if not (_X.dtype == np.float or _X.dtype == np.double):
        _X = _X.astype('category')
        _XEncoder = OneHotEncoder(drop='if_binary').fit(_X.values.codes.reshape(-1, 1))
    else:
        _XEncoder = None

    if not (_y.dtype == np.float or _y.dtype == np.double):
        _y = _y.astype('category')

    # prepare plots
    label_fig, label_ax = plt.subplots(cv.get_n_splits(), 1, figsize=(25, 10))
    label_fig.suptitle(_y.name)
    feature_lags_fig, feature_lags_ax = plt.subplots(1, 1 if _X.dtype.name == 'category' else 2, figsize=(25, 10))

    # estimate feature quality for each train test set and lag
    lags = list(range(lags)) if isinstance(lags, int) else lags

    # execute the following 2 times where we use pct_change() the 2nd time
    _XX = [_X] if _X.dtype.name == 'category' else [_X, _X.pct_change()]
    for xi, _X in enumerate(_XX):
        scores = []

        for li, lag in enumerate(lags):
            # lag data
            lag_score = []
            X = _X.shift(lag)
            y = _y

            # sync data
            idx = X.dropna().index.intersection(y.dropna().index)
            X = X.loc[idx]
            y = y.loc[idx]

            # analyze feature quality
            for i, (train_index, test_index) in enumerate(cv.split(X)):
                train_y = y.iloc[train_index]
                test_y = y.iloc[test_index]

                # bring features in the expected shape and dummy code categorical variables
                if X.dtype.name == 'category':
                    encoded_x = _XEncoder.transform(X.values.codes.reshape(-1, 1)).toarray()
                else:
                    encoded_x = X.values.reshape(-1, 1)

                # add a constant trend such that we can regress vom trending to non trending vice versa
                encoded_x = np.concatenate([encoded_x, np.arange(len(encoded_x)).reshape(-1, 1)], axis=1)

                # perform linear or logistic regression
                if y.dtype.name == 'category':
                    if li == 0:
                        label_ax[i].scatter(train_y.index, train_y, c='b')
                        label_ax[i].scatter(test_y.index, test_y, c='r')

                    # One hot encode labels and perform regression
                    encoded_y = LabelEncoder().fit_transform(y.values.codes)
                    reg = classifier.fit(encoded_x[train_index], encoded_y[train_index])
                    y_hat = reg.predict(encoded_x[test_index])

                    # score model using f1
                    f1 = f1_score(encoded_y[test_index], y_hat, average=None).mean()
                    lag_score.append(f1)
                else:
                    if li == 0:
                        label_ax[i].plot(train_y, color='b')
                        label_ax[i].plot(test_y, color='r')

                    reg = regressor.fit(encoded_x[train_index], y.iloc[train_index].values)
                    y_hat = reg.predict(encoded_x[test_index])

                    # score model using r^2
                    r2 = r2_score(y.iloc[test_index], y_hat)
                    lag_score.append(r2)

            scores.append(lag_score)

        # add whisker plot for r^2 (or f1)
        feature_lags_fig.suptitle("f1" if _X.dtype.name == 'category' else 'r^2')
        ax = (feature_lags_ax[xi] if hasattr(feature_lags_ax, '__getitem__') else feature_lags_ax)
        ax.boxplot(scores, vert=False)
        ax.set_yticklabels([f'lag({l})' if xi == 0 else f'%(lag({l}))' for l in lags])
        ax.invert_yaxis()

    # return a summary of all plots and scores
    return FeatureScoreReport(label_fig, feature_lags_fig, scores)



# "^VIX", "Close"

#cv = TimeSeriesSplit(n_splits=5)
#cv = KFold(n_splits=5)
#lala(df, lambda df: df["^VIX", "Close"].shift(-1), [("^VIX", "Close")], cv)





from unittest import TestCase

from pandas_ml_utils.df.model import Model
from pandas_ml_common_test.config import TEST_DF


class TestDfModelExtension(TestCase):

    def test__property(self):
        self.assertIsInstance(TEST_DF.model, Model)




"""Augment pandas DataFrame with methods for machine learning"""
__version__ = '0.1.15'

import os
from setuptools import setup, find_packages


setup(
   name=os.path.basename(os.path.dirname(os.path.abspath(__file__))),
   version=__version__,
   author='KIC',
   author_email='',
   packages=find_packages(),
   scripts=[],
   url='https://github.com/KIC/pandas-ml-quant',
   license='MIT',
   description=__doc__,
   long_description=open('Readme.md').read(),
   long_description_content_type='text/markdown',
   install_requires=["pandas-ml-common", *open("requirements.txt").read().splitlines()],
   extras_require={
      "dev": open("dev-requirements.txt").read().splitlines(),
   },
   include_package_data=True,
   classifiers=[
      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
      'Development Status :: 3 - Alpha',
      'Intended Audience :: Developers',
      'Topic :: Software Development :: Build Tools',
      'License :: OSI Approved :: MIT License',
      'Programming Language :: Python :: 3',
      'Programming Language :: Python :: 3.7',
   ],
   keywords=['pandas', 'ml', 'util', 'quant'],
)



import os
import tempfile
import uuid

from pandas_ml_common import pd, np
from pandas_ml_utils import FeaturesAndLabels, Model, AutoEncoderModel
from pandas_ml_utils.ml.data.splitting import RandomSplits, NaiveSplitter


class TestAbstractModel(object):

    def test_classifier(self):
        """given some toy classification data"""
        df = pd.DataFrame({
            "a": [1, 0, 1, 0, 1, 0, 1, 0,],
            "b": [0, 0, 1, 1, 0, 0, 1, 1,],
            "c": [1, 0, 0, 1, 1, 0, 0, 1,]
        })

        """and a model"""
        model = self.provide_classification_model(FeaturesAndLabels(features=["a", "b"], labels=["c"], label_type=int))

        """when we fit the model"""
        fit = df.model.fit(model, NaiveSplitter(0.49), verbose=0, epochs=1500)
        print(fit.training_summary.df)

        prediction = df.model.predict(fit.model)
        binary_prediction = prediction.iloc[:,0] >= 0.5
        np.testing.assert_array_equal(binary_prediction, np.array([True, False, False, True, True, False, False, True,]))

        """and save and load the model"""
        temp = os.path.join(tempfile.gettempdir(), str(uuid.uuid4()))
        try:
            fit.model.save(temp)
            copy = Model.load(temp)
            pd.testing.assert_frame_equal(df.model.predict(fit.model), df.model.predict(copy), check_less_precise=True)
        finally:
            os.remove(temp)

    def test_regressor(self):
        """given some toy regression data"""
        df = pd.DataFrame({
            "a": [-1.0, 0.0, 1.0, 2.0, 3.0, 4.0],
            "b": [-2.0, 1.0, 4.0, 7.0, 10.0, 13.0]
        })

        """and a model"""
        model = self.provide_regression_model(FeaturesAndLabels(features=["a"], labels=["b"]))

        """when we fit the model"""
        fit = df.model.fit(model, RandomSplits(0.3), verbose=0, epochs=500)
        print(fit.training_summary.df)

        """then we can predict"""
        prediction = df.model.predict(fit.model)
        np.testing.assert_array_almost_equal(prediction.iloc[:, 0].values, df["b"].values, 1)

        """and save and load the model"""
        temp = os.path.join(tempfile.gettempdir(), str(uuid.uuid4()))
        try:
            fit.model.save(temp)
            copy = Model.load(temp)
            pd.testing.assert_frame_equal(df.model.predict(fit.model), df.model.predict(copy), check_less_precise=True)
        finally:
            os.remove(temp)

    def test_auto_encoder(self):
        """given the implementation can handle auto encoders"""
        model = self.provide_auto_encoder_model(FeaturesAndLabels(["a", "b"], ["a", "b"]))
        if model is None:
            return

        """and some toy classification data"""
        df = pd.DataFrame({
            "a": [1, 0, 1, 0, ],
            "b": [0, 1, 0, 1, ],
        })

        """when we fit the model"""
        fit = df.model.fit(model, NaiveSplitter(0.49), verbose=0, epochs=500)
        print(fit.training_summary.df)

        """then we can encoder"""
        encoded_prediction = df.model.predict(fit.model.as_encoder())
        print(encoded_prediction)

        """and we can decoder"""
        decoder_features = encoded_prediction.columns.to_list()[0:1]
        decoded_prediction = encoded_prediction.model.predict(fit.model.as_decoder(decoder_features))
        print(decoded_prediction)
        np.testing.assert_array_almost_equal(decoded_prediction["prediction"].values > 0.5,
                                             df[["a", "b"]].values)

        """and we can encoder and decore after safe and load"""
        temp = os.path.join(tempfile.gettempdir(), str(uuid.uuid4()))
        try:
            fit.model.save(temp)
            copy = Model.load(temp)

            pd.testing.assert_frame_equal(
                df.model.predict(fit.model.as_encoder()),
                df.model.predict(copy.as_encoder()),
                check_less_precise=True)

            pd.testing.assert_frame_equal(
                encoded_prediction.model.predict(fit.model.as_decoder(decoder_features)),
                encoded_prediction.model.predict(copy.as_decoder(decoder_features)),
                check_less_precise=True)
        finally:
            os.remove(temp)

        # try to save only as encoder model
        try:
            fit.model.as_encoder().save(temp)
            copy = Model.load(temp)
        finally:
            os.remove(temp)

    # Abstract methods
    def provide_regression_model(self, features_and_labels) -> Model:
        pass

    def provide_classification_model(self, features_and_labels) -> Model:
        pass

    def provide_auto_encoder_model(self, features_and_labels) -> AutoEncoderModel:
        return None


from __future__ import annotations

import logging
from time import perf_counter
from typing import Callable, Dict, TYPE_CHECKING

import pandas as pd

from pandas_ml_common.utils import merge_kwargs, to_pandas
from pandas_ml_utils.constants import *
from pandas_ml_utils.ml.data.extraction import extract_feature_labels_weights, extract, extract_features
from pandas_ml_utils.ml.data.reconstruction import assemble_prediction_frame
from pandas_ml_utils.ml.data.splitting import DummySplitter, Splitter
from pandas_ml_utils.ml.data.splitting.random_splits import RandomSplits
from pandas_ml_utils.ml.data.splitting.sampeling import DataGenerator
from pandas_ml_utils.ml.fitting.fit import Fit
from pandas_ml_utils.ml.model import Model
from pandas_ml_utils.ml.summary import Summary
from .exception import FitException

_log = logging.getLogger(__name__)

if TYPE_CHECKING:
    pass


def fit(df: pd.DataFrame,
        model_provider: Callable[[int], Model],
        training_data_splitter: Splitter = RandomSplits(),
        hyper_parameter_space: Dict = None,
        **kwargs
        ) -> Fit:
    """

    :param df: the DataFrame you apply this function to
    :param model_provider: a callable which provides a new :class:`.Model` instance i.e. for each hyper parameter if
           hyper parameter tuning is enforced. Usually all the Model subclasses implement __call__ thus they are a
           provider of itself
    :param training_data_splitter: a :class:`pandas_ml_utils.ml.data.splitting.Splitter` object
           which provides traning and test data splits (eventually multiple folds)
    :param hyper_parameter_space: space of hyper parameters passed as kwargs to your model provider
    :return: returns a :class:`pandas_ml_utils.model.fitting.fit.Fit` object
    """

    trails = None
    model = model_provider()
    kwargs = merge_kwargs(model.features_and_labels.kwargs, model.kwargs, kwargs)
    (features, min_required_samples), labels, targets, weights, gross_loss = \
        extract(model.features_and_labels, df, extract_feature_labels_weights, **kwargs)

    start_performance_count = perf_counter()
    _log.info("create model")

    # get indices and make training and test data sets
    #train_idx, test_idx = training_data_splitter.train_test_split(features.index)
    #train = (features.loc[train_idx], labels.loc[train_idx], loc_if_not_none(weights, train_idx))
    #test = (features.loc[test_idx], labels.loc[test_idx], loc_if_not_none(weights, test_idx))

    # FIXME eventually perform a hyper parameter optimization first
    #if hyper_parameter_space is not None:
    #    # next isolate hyperopt parameters and constants only used for hyper parameter tuning like early stopping
    #    constants = {}
    #    hyperopt_params = {}
    #    for k, v in list(hyper_parameter_space.items()):
    #        if k.startswith("__"):
    #            hyperopt_params[k[2:]] = hyper_parameter_space.pop(k)
    #        elif isinstance(v, (int, float, bool)):
    #            constants[k] = hyper_parameter_space.pop(k)
    #
    #    # optimize hyper parameters
    #    model, trails = __hyper_opt(hyper_parameter_space,
    #                                hyperopt_params,
    #                                constants,
    #                                model_provider,
    #                                None, # FIXME Ecross_validation,
    #                                train,
    #                                test)

    # finally train the model with eventually tuned hyper parameters
    sampler = DataGenerator(training_data_splitter, features, labels, targets, weights, gross_loss).train_test_sampler()
    model.fit(sampler, **kwargs)
    _log.info(f"fitting model done in {perf_counter() - start_performance_count: .2f} sec!")

    # assemble result objects
    train_sampler, train_idx = sampler.training()
    test_sampler, test_idx = sampler.validation()

    try:
        prediction = (to_pandas(model.predict(train_sampler, **kwargs), train_idx, labels.columns),
                      to_pandas(model.predict(test_sampler, **kwargs), test_idx, labels.columns))

        # get training and test data tuples of the provided frames
        features, labels, targets, weights, gross_loss = sampler[0], sampler[1], sampler[2], sampler[3], sampler[4]
        df_train, df_test = [
            _assemble_result_frame(targets[i], prediction[i], labels[i], gross_loss[i], weights[i], features[i])
            for i in range(2)]

        # update model properties and return the fit
        model._validation_indices = test_idx
        model.features_and_labels.set_min_required_samples(min_required_samples)
        model.features_and_labels.set_label_columns(labels[0].columns.tolist())
        model.features_and_labels._kwargs = {k: a for k, a in kwargs.items() if k in model.features_and_labels.kwargs}

        return Fit(model, model.summary_provider(df_train, model, **kwargs), model.summary_provider(df_test, model, **kwargs), trails, **kwargs)
    except Exception as e:
        raise FitException(e, model)


def predict(df: pd.DataFrame, model: Model, tail: int = None, samples: int = 1, **kwargs) -> pd.DataFrame:
    min_required_samples = model.features_and_labels.min_required_samples

    if tail is not None:
        if min_required_samples is not None:
            # just use the tail for feature engineering
            df = df[-(abs(tail) + (min_required_samples - 1)):]
        else:
            _log.warning("could not determine the minimum required data from the model")

    kwargs = merge_kwargs(model.features_and_labels.kwargs, model.kwargs, kwargs)
    columns, features, targets = extract(model.features_and_labels, df, extract_features, **kwargs)

    if samples > 1:
        print(f"draw {samples} samples")

    sampler = DataGenerator(DummySplitter(samples), features, None, targets, None).complete_samples()
    predictions = model.predict(sampler, **kwargs)

    y_hat = to_pandas(predictions, index=features.index, columns=columns)
    return _assemble_result_frame(targets, y_hat, None, None, None, features)


def backtest(df: pd.DataFrame, model: Model, summary_provider: Callable[[pd.DataFrame], Summary] = None, **kwargs) -> Summary:
    kwargs = merge_kwargs(model.features_and_labels.kwargs, model.kwargs, kwargs)
    (features, _), labels, targets, weights, gross_loss =\
        extract(model.features_and_labels, df, extract_feature_labels_weights, **kwargs)

    sampler = DataGenerator(DummySplitter(1), features, labels, targets, None).complete_samples()
    predictions = model.predict(sampler, **kwargs)

    y_hat = to_pandas(predictions, index=features.index, columns=labels.columns)
    df_backtest = _assemble_result_frame(targets, y_hat, labels, gross_loss, weights, features)
    return (summary_provider or model.summary_provider)(df_backtest, model, **kwargs)


def _assemble_result_frame(targets, prediction, labels, gross_loss, weights, features):
    return assemble_prediction_frame(
        {TARGET_COLUMN_NAME: targets,
         PREDICTION_COLUMN_NAME: prediction,
         LABEL_COLUMN_NAME: labels,
         GROSS_LOSS_COLUMN_NAME: gross_loss,
         SAMPLE_WEIGHTS_COLUMN_NAME: weights,
         FEATURE_COLUMN_NAME: features})



"""
@ignore_warnings(category=ConvergenceWarning)
def __hyper_opt(hyper_parameter_space,
                hyperopt_params,
                constants,
                model_provider,
                cross_validation,
                train,
                test):
    from hyperopt import fmin, tpe, Trials

    keys = list(hyper_parameter_space.keys())

    def f(args):
        sampled_parameters = {k: args[i] for i, k in enumerate(keys)}
        model = None # FIXME model_provider(**join_kwargs(sampled_parameters, constants))
        loss = None # FIXME __train_loop(model, cross_validation, train, test)
        if loss is None:
            raise ValueError("Can not hyper tune if model loss is None")

        return {'status': 'ok', 'loss': loss, 'parameter': sampled_parameters}

    trails = Trials()
    fmin(f, list(hyper_parameter_space.values()), algo=tpe.suggest, trials=trails, show_progressbar=False, **hyperopt_params)

    # find the best parameters amd make sure to NOT pass the constants as they are only used for hyperopt
    best_parameters = trails.best_trial['result']['parameter']
    best_model = model_provider(**best_parameters)

    print(f'best parameters: {repr(best_parameters)}')
    return best_model, trails

"""


from unittest import TestCase

from pandas_ml_utils.ml.data.selection.feature_selection import score_feature
from test.config import DF_TEST


class TestFeatureSelection(TestCase):

    def test_continuous_feature_continuous_label(self):
        score = score_feature(DF_TEST, "Close", "Open", [1, 2, 3])
        print(score.scores)

    def test_discrete_feature_continuous_label(self):
        df = DF_TEST.copy()
        df["discrete"] = (df["Open"] / df["Close"] * 10).round().astype(int)
        score = score_feature(df, "Close", "discrete", [1, 2, 3])
        print(score.scores)

    def test_continuous_feature_discrete_label(self):
        df = DF_TEST.copy()
        df["discrete"] = (df["Open"] / df["Close"] * 10).round().astype(int)
        score = score_feature(df, "discrete", "Close", [1, 2, 3])
        print(score.scores)

    def test_discrete_feature_discrete_label(self):
        df = DF_TEST.copy()
        df["discrete"] = (df["Open"] / df["Close"] * 10).round().astype(int)
        score = score_feature(df, "discrete", "discrete", [1, 2, 3])
        print(score.scores)



from typing import Dict

# make sure to import monkey patched data frame
from pandas_ml_common import pd, np, get_pandas_object
from pandas_ml_utils.constants import PREDICTION_COLUMN_NAME, TARGET_COLUMN_NAME


def assemble_prediction_frame(frames: Dict[str, pd.DataFrame]):
    # filter non frames
    valid_frames = {head: frame.copy() for head, frame in frames.items() if frame is not None}

    for head, frame in valid_frames.items():
        frame.columns = pd.MultiIndex.from_product([[head], frame.columns.to_list()])

    # join all frames and keep the order of the passed dictionary
    df = pd.concat(valid_frames.values(), axis=1, join='inner', copy=False)

    # monkey patch prediction frame
    df.map_prediction_to_target = lambda: map_prediction_to_target(df, PREDICTION_COLUMN_NAME, TARGET_COLUMN_NAME)
    return df


def map_prediction_to_target(df, prediction, targets):
    def _round(val, d):
        return round(val, d) if isinstance(val, float) else val

    dfp = get_pandas_object(df, prediction)
    p = dfp._.values.reshape((len(df), -1))

    dft = get_pandas_object(df, targets)
    t = dft._.values.reshape((len(df), -1))

    if p.shape[1] == t.shape[1]:
        # 1:1 mapping
        index = [(date, _round(target, 2)) for date in df.index for target in dft.loc[date].values]
    elif p.shape[1] == t.shape[1] - 1:
        # we need to build ranges
        def build_tuples(l):
            return [(_round(l[i - 1], 2), _round(l[i], 2)) for i in range(1, len(l))]

        index = [(date, f"{target}") for date in df.index for target in
                 build_tuples(dft.loc[date].tolist())]
    elif p.shape[1] == t.shape[1] + 1:
        # mapping of the left and right extremes using +/- inf
        def build_tuples(l):
            l = [-np.inf, *l, np.inf]
            return [(_round(l[i - 1], 2), _round(l[i], 2)) for i in range(1, len(l))]

        index = [(date, target) for date in df.index for target in
                 build_tuples(dft.loc[date]._.values.tolist())]
    else:
        raise ValueError(f"unable to match {p.shape[1]} predictions to {t.shape[1]} +/-1 targets")

    return pd.DataFrame({"prediction": p.reshape((-1,))},
                        index=pd.MultiIndex.from_tuples(index))

